

/*****************************************************************/
/* bliksem.c is the main file of the bliksem theorem prover.     */
/* Bliksem 1.10                                                  */
/*****************************************************************/


#define CLASS 
#include "basics.h"
#include "extfunc.h" 
#include "resolution.h"


#define SHOWFILTERS    0

/****************************************************************/
/*               Error functions:                               */
/****************************************************************/


/* Since the program is non-interactive, there is no recovery after */
/* an error.                                                        */

static void closefiles ( )
{
   int i;

   for( i = freeincludefiles - 1; i >= 0; i -- ) 
   {
      if( includefiles [i]. rw == FILE_READ )
      {
         fclose ( includefiles [i]. handle ); 
         includefiles [i]. rw = FILE_NOTINUSE;
      }   
   }
   if( inputfile. rw == FILE_READ )
   {
      fclose( inputfile. handle );
      inputfile. rw = FILE_NOTINUSE; 
   }
}

/* print s, followed by 'in line %d of file %f', and exit.   */

void errorwithline ( s )
char *s;
{
   int i;

   printf( "%s ", s );
  
   for( i = freeincludefiles - 1; i >= 0; i -- )
   {
      if( includefiles [i]. rw == FILE_READ ) 
      {
         printf( "in line %d of file %s\n", 
                 includefiles [i]. linenum,
                 includefiles [i]. name );
      }
   }

   if( inputfile. rw == FILE_READ ) 
   {
      printf("in line %d of file %s\n",
         inputfile. linenum, inputfile. name ); 
   }  
   printf("\n"); 
   closefiles ();
   printstatistics( &outputfile ); 
   exit(0); 
}


/* print s, and exit.                      */

void errorwithoutline ( s )
char *s;
{
   printf("%s", s );
   printf("\n\n");
   closefiles ();
   printstatistics( &outputfile ); 
   exit(0); 
}


void internalproblem( s )
char *s;
{  
   printf("\n\n");
   printf("!!! Internal Problem: %s !!!", s );
   printf("\n\n");
   closefiles( );
   exit(0);
}


/******************************************************************/
/*      Functions for printing clauses:                           */
/******************************************************************/


/* print a clause in bliksem format.           */

static void printbliksemclause( f, clause )
bliksemfile *f;
int clause;
{
   int c;

   printstring( f, "(");
   printint( f, clauses [ clause + CLAUSE_NUMBER ] );
   printstring( f, ") ");

#if SHOWFILTERS
   /* Print the subsumption filters: */

   printstring( f, "[ " );
   printhex( f, clauses [ clause + CLAUSE_FILT1 ] );
   printstring( f, " ");
   printhex( f, clauses [ clause + CLAUSE_FILT2 ] );
   printstring( f, " ");
   printhex( f, clauses [ clause + CLAUSE_FILT3 ] );
   printstring( f, " ");
   printhex( f, clauses [ clause + CLAUSE_FILT4 ] );
   printstring( f, " ");
   printhex( f, clauses [ clause + CLAUSE_FILT5 ] );
   printstring( f, " ");
   printhex( f, clauses [ clause + CLAUSE_FILT6 ] );
   printstring( f, " ");
   printhex( f, clauses [ clause + CLAUSE_FILT7 ] );
   printstring( f, " ] ");
#endif

   printstring( f, "{" ); 
   printstring( f, "G" );
      printint( f, clauses [ clause + CLAUSE_GENERATION ] );
   printstring( f, ",W" );
      printint( f, clauses [ clause + CLAUSE_WEIGHT ] ); 
   printstring( f, ",D" );
      printint( f, clauses [ clause + CLAUSE_DEPTH ] ); 
   printstring( f, ",L" );
      printint( f, clauses [ clause + CLAUSE_LENGTH ] ); 
   printstring( f, ",V" );
      printint( f, clauses [ clause + CLAUSE_NRVARS ] ); 
   printstring( f, ",M" );
      printint( f, clauses [ clause + CLAUSE_END ] - 
                   clauses [ clause + CLAUSE_MAXLITS ] ); 
   printstring( f, "} " );

   c = clauses [ clause + CLAUSE_DERIVATION ];
   while( c < clauses [ clause + CLAUSE_LITERALS ] )
   { 
      switch( clauses [c] )
      {
      case RULE_RES:
         printstring( f, "R" ); 
         goto parents2; 
      case RULE_PARAMOD:
         printstring( f, "P" ); 
         goto parents2;

      case RULE_EQREFL:
         printstring( f, "Q" );
         goto parents1; 
      case RULE_EQFACT:
         printstring( f, "E" ); 
         goto parents1;
      case RULE_FACTOR:
         printstring( f, "F" );
         goto parents1;
      case RULE_RESIMP:
         printstring( f, "S" );
         goto parents1;
      case RULE_INITIAL:
         printstring( f, "I" ); 
         goto parents0;

      case SIMP_SPLITPOS:
         printstring( f, "p" );
         goto parents0;    
      case SIMP_SPLITNEG:
         printstring( f, "n" );
         goto parents0; 
      case SIMP_DEMOD:
         printstring( f, "d" );
         goto parents1;
      case SIMP_RES:
         printstring( f, "r");
         goto parents1;
      case SIMP_FACTOR:
         printstring( f, "f" );
         goto parents0;
      case SIMP_EQREFL:
         printstring( f, "q" );
         goto parents0;
      case SIMP_SELF:
         printstring( f, "s" );
         goto parents0;
      default:   
         internalproblem("Unknown rule");
         exit(0); 
      }
   
   parents0:
      c += 1;
      goto next;
   parents1:
      printstring( f, "(" );
      printint( f, clauses [ clauses[c+1] + CLAUSE_NUMBER ] );
      printstring( f, ")" );
      c += 2;
      goto next;
   parents2:
      printstring( f, "(" );
      printint( f, clauses [ clauses [c+1] + CLAUSE_NUMBER ] );
      printstring( f, "," ); 
      printint( f, clauses [ clauses [c+2] + CLAUSE_NUMBER ] );
      printstring( f, ")" );
      c += 3;
      goto next;
   next:
      if( c < clauses [ clause + CLAUSE_LITERALS ] )
         printstring( f, ";" ); 
   }

   /* the literals: */

   printstring( f, " { " ); 
   for( c = clauses [ clause + CLAUSE_LITERALS ]; 
        c < clauses [ clause + CLAUSE_END ];
        c ++ )
   {
      printterm( f, clauses [c] );
      if( c + 1 != clauses [ clause + CLAUSE_END ] )
         printstring( f, ", " );
   }
   printstring( f, " }.\n" ); 
}



/* Print a clause in prolog readable format.                     */

static void printprologclause( f, clause )
bliksemfile *f;
int clause;
{
   int lit;
   int litstart, litend;

   printstring( f, "clause( " );
   printint( f, clauses [ clause + CLAUSE_NUMBER ] );
 
   litstart =  clauses [ clause + CLAUSE_LITERALS ];
   litend = clauses [ clause + CLAUSE_END ];

   printstring( f, ", " );
   if( litstart < litend )
   { 
      printstring( f, "[ " );
      for( lit = litstart; lit < litend; lit ++ )
      {
         printterm( f, clauses [ lit ] );
         if( lit < clauses [ clause + CLAUSE_END ] - 1 )
            printstring( &outputfile, ", " );
      }
      printstring( &outputfile, " ]" );
   }
   else
      printstring( &outputfile, "[]");
   printstring( &outputfile, " )\n");
}


/* Print a clause. Choose the format.    */

void printclause( f, clause )
bliksemfile *f;
int clause;
{
   if( prologoutput )
      printprologclause( f, clause );
   else
      printbliksemclause( f, clause);
}




void printclauselist( f, clause )
bliksemfile *f;
int clause;
{
   while( clause != EMPTYCLAUSELIST )
   {
      printclause( f, clause );
      clause = clauses [ clause + CLAUSE_NEXTINLIST ];

      if( prologoutput )
         printstring( f, ".\n" );
   }
}





/*************************************************************/
/*            Additional functions:                          */
/*************************************************************/

/* Try to parse an assignment of the form Set( blabla, Value ). */
/* The Set has already been recognized. Return 1 if the         */
/* parse succeeded.                                             */

static int tryassignment( term, s, x, minval, maxval )
int term; 
char *s;
int *x;
int minval;
int maxval;
{
   char *s1;
   int f;

   f = termspace [ term ];
   if( f < 0 || arities [f] != 2 )
      return 0;

   f = termspace [ term + 1 ];
   if( f < 0 || arities [f] != 0 )
      return 0;

   f = termspace [ term + 2 ];
   if( f < 0 || arities [f] != 0 )
      return 0;

   s1 = operatorstring( termspace [ term + 1 ] );
   if( strcmp( s, s1 ) != 0 )
      return 0;

   s1 = operatorstring( termspace [ term + 2 ] );
   sscanf( s1, "%d", x );
   if( *x < minval || *x > maxval )
   {
      printstring( &outputfile, "value for assignment is out of range:\n" );
      printterm( &outputfile, term );
      errorwithoutline( "" );
   }
   return 1;
}


/* Set the operator weight/order/associativity/select.        */
/* term should have form xxx( op(...), value ).               */
     
static void operatorwoasp( woasp, term, minval, maxval ) 
int woasp [];
int term;
int minval;
int maxval;
{
   int op;
   int value, sign;
   int arg2;
   char *s;

   op = termspace [ term ];
   if( arities [ op ] != 2 ) goto wrong;

   op = termspace [ term + 1 ];
   if( op < 0 || op >= freeoperators )
      internalproblem( "operatorwoasp" );

   arg2 = termends [ term + 1 ];
   if( termspace [ arg2 ] == OP_UNARYMINUS )
   {
      sign = -1;
      arg2 ++ ;
   }
   else
      sign = 1;
   if( arities [ termspace [ arg2 ]] != 0 ) goto wrong;

   s = operatorstring( termspace [ arg2 ] );
   sscanf( s, "%d", &value );
   value *= sign;

   if( value < minval || value > maxval )
   {
      printstring( &outputfile, "value for woasp is out of range:\n");
      printterm( &outputfile, term );
      errorwithoutline( "" );
   }
   woasp [ op ] = value;
   return;
wrong:
   printstring( &outputfile, "incorrect woasp assignment:\n");
   printterm( &outputfile, term );
   errorwithoutline( "" );
}


/* Try to parse term as an assignment of the form Set(litordering, bla). */

static int trylitordering( term )
int term;
{
   int op;
   int ord;
   char *s;

   op = termspace [ term ];
   if( arities [ op ] != 2 )
      return 0;
   op = termspace [ term + 1 ];
   if( op < 0 || arities [ op ] != 0 )
      return 0;
   s = operatorstring( op );
   if( strcmp( s, "litordering" ) != 0 )
      return 0;
   op = termspace [ term + 2 ];
   if( op < 0 || arities [ op ] != 0 )
      return 0;

   s = operatorstring( op );
   if( strcmp( s, "none" ) == 0 )
   {
      ord = ORD_NONE; goto found;
   } 
   if( strcmp( s, "decision" ) == 0 )
   {
      ord = ORD_DECISION; goto found;
   }
   if( strcmp( s, "exttermordpa" ) == 0 )
   {
      ord = ORD_EXTTERMORDPA; goto found;
   }
   if( strcmp( s, "exttermordap" ) == 0 )
   {
      ord = ORD_EXTTERMORDAP; goto found;
   }
   if( strcmp( s, "termordering" ) == 0 )
   {
      ord = ORD_TERMORDERING; goto found;
   }
   if( strcmp( s, "liftord" ) == 0 )
   {
      ord = ORD_LIFTORD; goto found;
   }
   if( strcmp( s, "nlo" ) == 0 )
   {
      ord = ORD_NLO; goto found;
   }
   if( strcmp( s, "split" ) == 0 )
   {
      ord = ORD_SPLIT; goto found;
   }
   
   printstring( &outputfile, "Did not recognize literal order:\n");
   printterm( &outputfile, term );
   errorwithoutline( " ");

found:
   if( nrlitorderings >= MAXLITORDERINGS - 2 ) 
      errorwithoutline( "too many literal orderings" );

   litorderings [ nrlitorderings ++ ] = ord;
   return 1;
}


/* Try to parse term as an expression of the form Set( substype, bla ). */

static int trysubstype( term )
int term;
{
   int op;
   char *s;

   op = termspace [ term ];
   if( arities [ op ] != 2 )
      return 0;
   op = termspace [ term + 1 ];
   if( op < 0 || arities [ op ] != 0 )
      return 0;
   s = operatorstring( op );
   if( strcmp( s, "substype" ) != 0 )
      return 0;
   op = termspace [ term + 2 ];
   if( op < 0 || arities [ op ] != 0 )
      return 0;

   s = operatorstring( op );
   if( strcmp( s, "standard" ) == 0 )
   {
      substype = SUBS_STANDARD; goto found;
   } 
   if( strcmp( s, "snl1" ) == 0 )
   {
      substype = SUBS_SNL1; goto found;
   }
   if( strcmp( s, "eqrewr" ) == 0 )
   {
      substype = SUBS_EQREWR; goto found;
   }
   
   printstring( &outputfile, "Did not recognize substype:\n");
   printterm( &outputfile, term );
   errorwithoutline( " ");

found:
   return 1;
}


static int tryconversion( term, fullconversion )
int term;
int *fullconversion;
{
   int op;
   char *s;

   op = termspace [ term ];
   if( arities [ op ] != 2 )
      return 0;
   op = termspace [ term + 1 ];
   if( op < 0 || arities [ op ] != 0 )
      return 0;
   s = operatorstring( op );
   if( strcmp( s, "conversion" ) != 0 )
      return 0;
   op = termspace [ term + 2 ];
   if( op < 0 || arities [ op ] != 0 )
      return 0;

   s = operatorstring( op );
   if( strcmp( s, "none" ) == 0 )
   {
      *fullconversion = FULLCONV_NONE;
      goto found; 
   } 
   if( strcmp( s, "standard" ) == 0 )
   {
      *fullconversion = FULLCONV_STANDARD;
      goto found;
   }
   if( strcmp( s, "guarded" ) == 0 )
   {
      *fullconversion = FULLCONV_GUARDED;
      goto found;
   }
   if( strcmp( s, "struct" ) == 0 )
   {
      *fullconversion = FULLCONV_STRUCT;
      goto found;
   }
   if( strcmp( s, "structos" ) == 0 )
   {
      *fullconversion = FULLCONV_STRUCTOS;
      goto found;
   }
   if( strcmp( s, "2var" ) == 0 )
   {
      *fullconversion = FULLCONV_2VAR;
      goto found;
   }
   
   printstring( &outputfile, "Did not recognize conversion:\n");
   printterm( &outputfile, term );
   errorwithoutline( " ");

found:
   printstring( &outputfile, "Selected conversion: ");
   printstring( &outputfile, s );
   printstring( &outputfile, "\n" );
   return 1;
}


/* Try to parse term as an expression of the form Set( termordering, bla ). */

static int trytermordering( term )
int term;
{
   int op;
   char *s;
   int sign;

   op = termspace [ term ];
   if( arities [ op ] != 2 )
      return 0;
   op = termspace [ term + 1 ];
   if( op < 0 || arities [ op ] != 0 )
      return 0;
   s = operatorstring( op );
   if( strcmp( s, "termordering" ) != 0 )
      return 0;
   op = termspace [ term + 2 ];

   if( op == OP_UNARYMINUS ) 
   {
      op = termspace [ term + 3 ] ;
      sign = -1;
   }
   else
      sign = 1;

   if( op < 0 || arities [ op ] != 0 )
      return 0;

   s = operatorstring( op );
   if( strcmp( s, "none" ) == 0 )
   {
      termordering = ORD_NONE; goto found;
   } 
   if( strcmp( s, "depth" ) == 0 )
   {
      termordering = ORD_DEPTH; goto found;
   }
   if( strcmp( s, "vardepth" ) == 0 )
   {
      termordering = ORD_VARDEPTH; goto found;
   }
   if( strcmp( s, "compl" ) == 0 )
   {
      termordering = ORD_COMPL; goto found;
   }
   if( strcmp( s, "nrvars" ) == 0 )
   {
      termordering = ORD_NRVARS; goto found;
   }
   if( strcmp( s, "vd" ) == 0 )
   {
      termordering = ORD_VD; goto found;
   }
   if( strcmp( s, "kbo" ) == 0 )
   {
      termordering = ORD_KBO; goto found;
   }
   if( strcmp( s, "lrpo" ) == 0 )
   {
      termordering = ORD_LRPO; goto found;
   }
   
   printstring( &outputfile, "Did not recognize term order:\n");
   printterm( &outputfile, term );
   errorwithoutline( "" );

found:
   termordering *= sign;
   return 1;
}


/* Try to parse term as an expression of the form Set( litselect, bla ). */

static int trylitselect( term )
int term;
{
   int op;
   char *s;

   op = termspace [ term ];
   if( arities [ op ] != 2 )
      return 0;
   op = termspace [ term + 1 ];
   if( op < 0 || arities [ op ] != 0 )
      return 0;
   s = operatorstring( op );
   if( strcmp( s, "litselect" ) != 0 )
      return 0;
   op = termspace [ term + 2 ];

   if( op < 0 || arities [ op ] != 0 )
      return 0;

   s = operatorstring( op );
   if( strcmp( s, "none" ) == 0 )
   {
      litselect = LITSEL_NONE; return 1;
   } 
   if( strcmp( s, "positive" ) == 0 )
   {
      litselect = LITSEL_POS; return 1;
   }
   if( strcmp( s, "negative" ) == 0 )
   {
      litselect = LITSEL_NEG; return 1;
   }
   if( strcmp( s, "posord" ) == 0 )
   {
      litselect = LITSEL_POSORD; return 1;
   }
   if( strcmp( s, "negord" ) == 0 )
   {
      litselect = LITSEL_NEGORD; return 1;
   }
   
   printstring( &outputfile, "Did not recognize literal selection:\n");
   printterm( &outputfile, term );
   errorwithoutline( "" );
}




static void printorder( f, order )
bliksemfile *f;
int order;
{
   int ord;

   if( order < 0 )
   {
      ord = -order;
      printstring( f, "-" );
   }
   else
      ord = order;
   
   switch( ord )
   {
   case ORD_NONE:
      printstring( f, "none" );
      break;
   case ORD_DEPTH:
      printstring( f, "depth" );
      break;   
   case ORD_VARDEPTH:
      printstring( f, "vardepth" );
      break;
   case ORD_COMPL:
      printstring( f, "complexity" );
      break;
   case ORD_NRVARS:
      printstring( f, "nrvars" );
      break;
   case ORD_VD:
      printstring( f, "vd" );
      break;
   case ORD_DECISION:
      printstring( f, "varnr + vardepth" );
      break;
   case ORD_KBO:
      printstring( f, "kbo" );
      break;
   case ORD_LRPO:
      printstring( f, "lrpo" );
      break;
   case ORD_EXTTERMORDPA:
      printstring( f, "extend the termordering, first sorting on predicates" );
      break;
   case ORD_EXTTERMORDAP:
      printstring( f, "extend the termordering, first sorting on arguments" );
      break;
   case ORD_TERMORDERING:
      printstring( f, "termordering" );
      break;
   case ORD_LIFTORD:
      printstring( f, "liftord" );
      break;
   case ORD_NLO:
      printstring( f, "nlo" );
      break;
   case ORD_SPLIT:
      printstring( f, "split" );
      break;
   default: 
      internalproblem("unknown order!" );
   }
}


static void printlitselect( f, litselect )
bliksemfile *f;
int litselect;
{
   switch( litselect )
   {
   case LITSEL_NONE:
      printstring( f, "none" );
      break;
   case LITSEL_POS:
      printstring( f, "positive" );
      break;
   case LITSEL_NEG:
      printstring( f, "negative" );
      break;
   case LITSEL_POSORD:
      printstring( f, "posord" );
      break;
   case LITSEL_NEGORD:
      printstring( f, "negord" );
      break;
   default:
      internalproblem( "unknown selection function" );
   }
}


 
static void printoperatorord( f, translation )
bliksemfile *f;
int translation;
{
   int s;
   int t0, t1, i;
   int nrprinted;

   printstring( f, "\n" );
   printstring( f, "Symbols occurring in the translation:\n\n" );

   t0 = translations [ translation ]. startofresult;
   t1 = translations [ translation ]. endofresult;
   nrprinted = 0;
 
   for( s = 0; s < freeoperators; s ++ )
   {
      for( i = t0; i < t1; i ++ )
      {
         if( termspace [i] == s || s == OP_REWRITES || s == OP_NOT || 
             s == OP_EQUAL )
         {
            if( nrprinted > 0 )
            {
               printstring( &outputfile, ", " );
               if( nrprinted % 1 == 0 )
                  printstring( &outputfile, "\n" );
            }
            printoperator( f, s );
            printstring( f, "  [" );
            printint( f, s );
            printstring( f, ", ");
            printint( f, arities [s] );
            printstring( f, "]      " );
            printstring( f, "(w:");
            printint( f, operatorweights [s] );
            printstring( f, ", o:");
            printint( f, operatororder [s] );
            printstring( f, ", a:");
            printint( f, operatorassoc [s] );
            printstring( f, ", s:");
            printint( f, operatorselect [s] ); 
            printstring( f, ", b:");
            printint( f, operatornoparamod [s] );
            printstring( f, ")" );

            nrprinted ++ ;
            goto printed;
         }
      }
   printed:
      ;
   }
   printstring( f, ".\n\n" );
}


static void printsubstype( f, substype )
bliksemfile *f;
int substype;
{
   switch( substype )
   {
   case SUBS_STANDARD:
      printstring( f, "standard" );
      break;
   case SUBS_SNL1:
      printstring( f, "snl1" );
      break;
   case SUBS_EQREWR:
      printstring( f, "eqrewr" );
      break;
   default:
      errorwithoutline("Unknown substype!");
      exit(0);
   }
}

 
static void printoptions( f, translation )
bliksemfile *f;
int translation;
{
   int i;

   printstring( f, "\n\n" );
   printstring( f, "Options Used:\n\n" ); 

   printstring( f, "useres =            " );
   printint( f, useres );
   printstring( f, "\n" );

   printstring( f, "useparamod =        " );
   printint( f, useparamod );
   printstring( f, "\n" );

   printstring( f, "useeqrefl =         " );
   printint( f, useeqrefl );
   printstring( f, "\n" );

   printstring( f, "useeqfact =         " );
   printint( f, useeqfact );
   printstring( f, "\n" );

   printstring( f, "usefactor =         " );
   printint( f, usefactor );
   printstring( f, "\n" );

   printstring( f, "usesimpsplitting =  " );
   printint( f, usesimpsplitting );
   printstring( f, "\n" ); 

   printstring( f, "usesimpdemod =      " );
   printint( f, usesimpdemod ); 
   printstring( f, "\n" ); 

   printstring( f, "usesimpres =        " );
   printint( f, usesimpres );
   printstring( f, "\n" ); 

   printstring( f, "\n" );

   printstring( f, "resimpinuse      =  " );
   printint( f, resimpinuse );
   printstring( f, "\n" );

   printstring( f, "resimpclauses =     " );
   printint( f, resimpclauses );
   printstring( f, "\n" );

   printstring( f, "substype =          " );
   printsubstype( f, substype );
   printstring( f, "\n" );

   printstring( f, "backwardsubs =      " );
   printint( f, backwardsubs );
   printstring( f, "\n" );

   printstring( f, "selectoldest =      " );
   printint( f, selectoldest );
   printstring( f, "\n" );

   if( nrlitorderings > 0 )
      printstring( f, "\n" ); 

   for( i = 0; i < nrlitorderings; i ++ )
   {
      printstring( f, "litorderings [");
      printint( f, i );
      printstring( f, "] =  " );
      printorder( f, litorderings [i] );
      printstring( f, "\n");

   }
   printstring( f, "\n" );

   printstring( f, "termordering =      " );
   printorder( f, termordering );
   printstring( f, "\n" );
   printstring( f, "\n" );

   printstring( f, "litapriori =        " );
   printint( f, litapriori );
   printstring( f, "\n" );

   printstring( f, "termapriori =       " );
   printint( f, termapriori );
   printstring( f, "\n" );

   printstring( f, "litaposteriori =    " );
   printint( f, litaposteriori );
   printstring( f, "\n" );

   printstring( f, "termaposteriori =   " );
   printint(f, termaposteriori );
   printstring( f, "\n" );

   printstring( f, "demodaposteriori =  " );
   printint( f, demodaposteriori );
   printstring( f, "\n" );

   printstring( f, "ordereqreflfact =   " );
   printint( f, ordereqreflfact );
   printstring( f, "\n\n" );

   printstring( f, "litselect =         " );
   printlitselect( f, litselect );
   printstring( f, "\n\n" );

   printstring( f, "maxweight =         " );
   printint( f, maxweight );
   printstring( f, "\n" );

   printstring( f, "maxdepth =          " );
   printint( f, maxdepth );
   printstring( f, "\n" );

   printstring( f, "maxlength =         " );
   printint( f, maxlength );
   printstring( f, "\n" );

   printstring( f, "maxnrvars =         " );
   printint( f, maxnrvars );
   printstring( f, "\n" );

   printstring( f, "excuselevel =       " );
   printint( f, excuselevel );
   printstring( f, "\n" );

   printstring( f, "increasemaxweight = " );
   printint( f, increasemaxweight );
   printstring( f, "\n\n" );

   printstring( f, "maxselected =       " );
   printint( f, maxselected );
   printstring( f, "\n" );

   printstring( f, "maxclauses =        " );
   printint( f, maxclauses );
   printstring( f, "\n\n" );
   
   printstring( f, "showgenerated =    " );
   printint( f, showgenerated );
   printstring( f, "\n" );

   printstring( f, "showkept =         " );
   printint( f, showkept );
   printstring( f, "\n" );

   printstring( f, "showselected =     " );
   printint( f, showselected );
   printstring( f, "\n" );

   printstring( f, "showdeleted =      " );
   printint( f, showdeleted );
   printstring( f, "\n" );

   printstring( f, "showresimp =       " );
   printint( f, showresimp );
   printstring( f, "\n" );

   printstring( f, "showstatus =       " );
   printint( f, showstatus );
   printstring( f, "\n");
   printstring( f, "\n" );

   printstring( f, "prologoutput =     " );
   printint( f, prologoutput );
   printstring( f, "\n" );

   printstring( f, "nrgoals =          " );
   printint( f, nrgoals );
   printstring( f, "\n" );

   printstring( f, "totalproof =       " );
   printint( f, totalproof );
   printstring( f, "\n" );

   printoperatorord( f, translation );

}


/****************************************************************/
/*   Test membership in the decidable classes.                  */
/****************************************************************/

/* Return 1 if every non-ground, functional term in literal contains */
/* at least nrvar variables.                                         */

static int weaklycovers( literal, nrvars )
int literal;
int nrvars;
{
   int term, atom;
   int end;
   int nv;

   if( termspace [ literal ] == OP_NOT ) 
      atom = literal + 1;
   else
      atom = literal;
   end = termends [ atom ];
   term = atom + 1;

   while( term < end )
   {
      if( termspace [ term ] >= 0 )
      {
         nv = termnrvars( term );
         if( nv != 0 && nv != nrvars )
            return 0;
      }
      term ++ ;
   }
   return 1;
}



/* Return 1 if the clauselist starting in clause is guarded.   */

static int listguarded( clause )
int clause;
{
   int cl, term;
   int nrvars, hasguard;

   while( clause != EMPTYCLAUSELIST )
   {
      nrvars = clausenrvars( clause );
      if( nrvars >= 2 )
      {
         hasguard = 0;
         for( cl = clauses [ clause + CLAUSE_LITERALS ];
              cl < clauses [ clause + CLAUSE_END ];
              cl ++ )
         {
            term = clauses [ cl ];
            if( ! weaklycovers( term, nrvars ))
               return 0;
            if( termspace [ term ] == OP_NOT ) 
            { 
               term ++ ;
               if( termnrvars( term ) == nrvars && termvardepth( term ) == 1 )
                  hasguard = 1;
            }
         }
         if( hasguard == 0 )
            return 0;
      } 
      clause = clauses [ clause + CLAUSE_NEXTINLIST ];
   }
   return 1;
} 


/* Return 1 if the clauselist, starting with clause is in E-plus.   */

static int listeplus( clause )
int clause;
{
   int cl, term;
   int nv, nrvars;

   while( clause != EMPTYCLAUSELIST )
   {
      nrvars = clausenrvars( clause );
      for( cl = clauses [ clause + CLAUSE_LITERALS ];
           cl < clauses [ clause + CLAUSE_END ];
           cl ++ )
      {
         term = clauses [ cl ];
         if( ! weaklycovers( term, nrvars ))
            return 0;
         nv = termnrvars( term );
         if( nv != 0 && nv != nrvars )
            return 0;
      }
      clause = clauses [ clause + CLAUSE_NEXTINLIST ];
   }
   return 1;         
}


/* Return 1 if the clauselist, starting with clause is in S-plus.   */

static int listsplus( clause )
int clause;
{
   int cl, term;
   int nv, nrvars;

   while( clause != EMPTYCLAUSELIST )
   {
      nrvars = clausenrvars( clause );
      for( cl = clauses [ clause + CLAUSE_LITERALS ];
           cl < clauses [ clause + CLAUSE_END ];
           cl ++ )
      {
         term = clauses [ cl ];
         if( ! weaklycovers( term, nrvars ))
            return 0;
         nv = termnrvars( term );
         if( nv >= 2 && nv != nrvars )
            return 0;
      }
      clause = clauses [ clause + CLAUSE_NEXTINLIST ];
   }
   return 1;         
}




/***************************************************************/
/*   Functions belonging to the automode.                      */
/***************************************************************/

/* Give the percentage of good Horn clauses.   */

static double pcthorn( clause )
int clause;
{
   int total, horn;
   int nrpos;
   int cl, lit;

   total = 0;
   horn = 0;
   while( clause != EMPTYCLAUSELIST ) 
   {
      nrpos = 0;
      for( cl = clauses [ clause + CLAUSE_LITERALS ];
           cl < clauses [ clause + CLAUSE_END ];
           cl ++ )
      {
         lit = clauses [ cl ];
         if( termspace [ lit ] != OP_NOT )
            nrpos ++ ;
      }
      if( nrpos <= 1 )
         horn ++ ;
      total ++ ; 

      clause = clauses [ clause + CLAUSE_NEXTINLIST ];
   }
   if( total != 0 )
      return ( (double) horn ) / ( (double) total ); 
   else
      return 1.0;
}


/* Produce pct. of equality literals in list.     */

static double pctequality( list )
int list;
{
   int nrliterals;
   int nrequality;
   int cl, at;

   nrliterals = 0;
   nrequality = 0;
   while( list != EMPTYCLAUSELIST ) 
   {
      for( cl = clauses [ list + CLAUSE_LITERALS ];
           cl < clauses [ list + CLAUSE_END ];
           cl ++ )
      {
         at = clauses [ cl ];
         if( termspace [ at ] == OP_NOT )
            at ++ ;
         if( termspace [ at ] == OP_EQUAL || termspace [ at ] == OP_REWRITES )
            nrequality ++ ;
         nrliterals ++;
      }
      list = clauses [ list + CLAUSE_NEXTINLIST ];
   }
   if( nrliterals > 0 )
      return (double) nrequality / (double) nrliterals;
   else
      return 0.0; 
}
 

/* Give standard woasp settings to the names introduced by the  */
/* normal form transformation.                                  */

static void woaspskolem( )
{
   int i;

   for( i = 0; i < freeoperators; i ++ )
   {
      if( (operators [i]. flags & FLAG_SUBFORM) != 0 ||
          (operators [i]. flags & FLAG_SKOLEM) != 0 )
      {
         operatorweights [i] = 1;
         operatororder [i] = i;
         operatorassoc [i] = 0;
         operatorselect [i] = 1;
         operatornoparamod [i] = 1;
      } 
   }
}


/* Give all symbols equal weight, with the exception of not and equality.  */

static void weightsequal( )
{
   int i;

   for( i = 0; i < freeoperators; i ++ )
      operatorweights [i] = 1;
   operatorweights [ OP_NOT ] = 0;   
}


/* Make all symbols in selectable by the selection function.   */

static void selectall( )
{
   int i;

   for( i = 0; i < freeoperators; i ++ )
      operatorselect [i] = 1;   
}



/* Return 1 if symbol s1 should be put before s2 in the order.   */

static int shouldbeless( s1, s2 )
int s1, s2;
{
   int cmp;

   if( arities [ s1 ] < arities [ s2 ] ) return 1;
   if( arities [ s1 ] > arities [ s2 ] ) return 0; 

   if( s1 < OP_NONPRIMITIVE && s2 >= OP_NONPRIMITIVE ) return 1;
   if( s2 < OP_NONPRIMITIVE ) return 0;

   cmp = strcmp( operatorstring( s1 ), operatorstring( s2 ));
   if( cmp == -1 ) return 1;
   if( cmp == 1 ) return 0;
   return 0;
}


/* Produce an order on the symbols, using shouldbeless.           */

static void ordersymbols( )
{
   int cnt, i;
   int best;

   for( i = 0; i < freeoperators; i ++ )
   {
      operatororder [i] = -1;
      if( termordering == ORD_LRPO )
         operatorassoc [i] = 1;
      else
         operatorassoc [i] = 1;
   }

   operatororder [ OP_EQUAL ] = 0; 
   operatororder [ OP_REWRITES ] = 0;
   operatorassoc [ OP_EQUAL ] = 0;
   operatorassoc [ OP_REWRITES ] = 0;

   for( cnt = 2; cnt < freeoperators; cnt ++ )
   {
      best = -1;
      for( i = 0; i < freeoperators; i ++ )
      {
         if( operatororder [i] == -1 )
         {
            if( best == -1 ||
               ( best >= 0 && best < freeoperators && shouldbeless( i, best)))
               best = i;
         }
      }
      if( best == -1 ) internalproblem( "could not find next in order" );
      operatororder [ best ] = cnt;
   }
}



static void noparamodskolemstruct( )
{
   int i;

   for( i = 0; i < freeoperators; i ++ )
   {
      if( ( operators [i]. flags & FLAG_SUBFORM ) != 0 ||
          ( operators [i]. flags & FLAG_SKOLEM  ) != 0 )
      {
         operatornoparamod [i] = 1;   
      } 
   }
}



static void settingsdecision( )
{
   useres = 1;
   useparamod = 0;
   useeqrefl = 0;
   useeqfact = 0;
   usefactor = 1;

   usesimpsplitting = 0;
   usesimpdemod = 0;
   usesimpres = 0;

   resimpinuse = 1000;
   resimpclauses = 10000;
   
   substype = SUBS_SNL1;
   selectoldest = 3;

   weightsequal( );
   selectall( );
   ordersymbols( );

   litapriori = 1;
   termapriori = ORD_NONE;
   termaposteriori = 0;
   litaposteriori = 0;
   demodaposteriori = 0;

   litorderings [0] = ORD_SPLIT;
   litorderings [1] = ORD_DECISION;
   litorderings [2] = ORD_EXTTERMORDPA;
   nrlitorderings = 3;

   incomplete = 0;
}


static void settingshorn( )
{
   useres = 1;
   useparamod = 0;
   usefactor = 1;
   useeqrefl = 0;
   useeqfact = 0;

   usesimpsplitting = 0;
   usesimpdemod = 0;
   usesimpres = 3;

   substype = SUBS_STANDARD;

   selectoldest = 3;
   litorderings [0] = ORD_SPLIT;
   litorderings [1] = ORD_LIFTORD;
   nrlitorderings = 2;

   litselect = LITSEL_NEG;
   
   termordering = ORD_NONE;
   litapriori = 1;
   termapriori = 0;
   litaposteriori = 0;
   termaposteriori = 0;
   demodaposteriori = 0;

   weightsequal( );
   operatorweights [ OP_NOT ] = 1;
   selectall( );
   ordersymbols( );

   increasemaxweight = 1;
}


static void settingsnlo( )
{
   useres = 1;
   useparamod = 0;
   
   useeqrefl = 0;
   useeqfact = 0;

   usefactor = 1;

   usesimpsplitting = 0;
   usesimpdemod = 0;
   usesimpres = 3;

   substype = SUBS_SNL1;
   weightsequal( );
   ordersymbols( );
   selectall( );

   litorderings [0] = ORD_SPLIT;
   litorderings [1] = ORD_NLO;
   nrlitorderings = 2;

   litapriori = 1;
   termapriori = 0;
   litaposteriori = 0;
   termaposteriori = 0;
   demodaposteriori = 0;
 
   excuselevel = 1;
   increasemaxweight = 1;
   maxweight = 15;
}



static void settingssomeequality( )
{
   useres = 1;
   useparamod = 1;

   useeqrefl = 1;
   useeqfact = 1;
   usefactor = 1;

   usesimpres = 3;
   usesimpdemod = 3;

   substype = SUBS_EQREWR;
   selectoldest = 3;

   litorderings [0] = ORD_SPLIT;
   litorderings [1] = ORD_EXTTERMORDAP;
   nrlitorderings = 2;

   termordering = ORD_KBO;

   litselect = LITSEL_NEGORD;

   termapriori = 1;
   litapriori = 1;
   termaposteriori = 1;
   litaposteriori = 1;
   demodaposteriori = 1;

   weightsequal( );
   ordersymbols( );
   selectall( );
   noparamodskolemstruct( );

   excuselevel = 1; 
   increasemaxweight = 1;
}


/* Make strategy selection. Return the conversion that should be used. */

static int automode( )
{
   int convstandard, convstructos, convguarded, conv2var;
   int i, strat;
   int freetermspace2, freeclauses2;
   double eq, horn;

   printstring( &outputfile, "Entering Automode\n\n" );

   /* First we construct a lot of clause transformations.   */

   nrtranslations = 0;

   convstandard = 
      makeconversion( startofinput, endofinput, FULLCONV_STANDARD );
   printstring( &outputfile, "\n" );

   convstructos = 
      makeconversion( startofinput, endofinput, FULLCONV_STRUCTOS );
   printstring( &outputfile, "\n" );

   convguarded = 
      makeconversion( startofinput, endofinput, FULLCONV_GUARDED );
   printstring( &outputfile, "\n" );

   conv2var = 
      makeconversion( startofinput, endofinput, FULLCONV_2VAR );
   printstring( &outputfile, "\n" );

   freetermspace2 = freetermspace;
   freeclauses2 = freeclauses;

   /* Try 3 decidable classes:       */   

   for( strat = 0; strat < 3; strat ++ )
   {
      usesimpsplitting = 0;
      usesimpdemod = 0;
      usesimpres = 0;

      switch( strat )
      {
      case 0: 
         (void) mainloop( convguarded, 1 );
         if( listguarded( listclauses ) && pctequality( listclauses ) == 0.0 )
         {
            printstring( &outputfile, "This is a guarded clause set\n\n");
            settingsdecision( );
            freetermspace = freetermspace2;
            freeclauses = freeclauses2;
            return convguarded;
         }
         break;
      case 1:
         (void) mainloop( convstandard, 1 );
         if( listeplus( listclauses ) && pctequality( listclauses ) == 0.0 )
         {
            printstring( &outputfile, "The clause set is in E-plus\n\n");
            settingsdecision( );
            freetermspace = freetermspace2;
            freeclauses = freeclauses2;
            return convstandard;
         }
         break;
      case 2:
         (void) mainloop( conv2var, 1 );
         if( listsplus( listclauses ) && pctequality( listclauses ) == 0.0 )
         {
            printstring( &outputfile, "The clause set is in S-plus\n\n" );
            settingsdecision( );
            freetermspace = freetermspace2;
            freeclauses = freeclauses2;
            return conv2var;
         }
         break;
      default:
         internalproblem( "3 should not happen" );
      }

      freetermspace = freetermspace2;
      freeclauses = freeclauses2;
   }

   /* We now have to make hard decisions:      */

   i = mainloop( convstructos, 1 );

   eq = pctequality( listclauses );
   horn = pcthorn( listclauses );
#if 1
   printf("eq = %lf, horn = %lf\n", eq, horn );
#endif

   if( eq == 0.0 ) goto noequality;
   if( eq >= 0.9 ) goto pureequality;

   goto someequality;

noequality:
   if( horn >= 0.75 )
   {
      printstring( &outputfile, 
         "This is a near-Horn, non-equality  problem\n" );
      settingshorn( );
   }
   else
   {
      printstring( &outputfile, "This a non-horn, non-equality problem\n" );
      settingsnlo( );
   }
   freetermspace = freetermspace2;
   freeclauses = freeclauses2; 
   return convstructos;

someequality:
   printstring( &outputfile, "This is a problem with some equality\n\n");

   freetermspace = freetermspace2;
   freeclauses = freeclauses2;
   settingssomeequality( ); 
   return convstructos;

pureequality:
   printstring( &outputfile, "This is a pure equality problem\n\n");
   freetermspace = freetermspace2;
   freeclauses = freeclauses2;
   settingssomeequality( );
   return convstructos;
}


/************************************************************/
/*                       main                               */
/************************************************************/



int main( argc, argv )
int argc;
char *argv [];
{
   int i; 
   int tt;
   char *s;
   int freetermspace2, freeclauses2; 
   int fullconversion;
   int convused;

   if( sizeof(int) < 4 )
   {
      printf("\nSorry, need at least 32 bit integers!\n\n");
      return 1;
   }

   initinoutput( argc, argv );

   freetermspace = 0; 
   freeclauses = 0; 
   freejustifications = 0;

   (void) parser ( );

   printstring( &outputfile, "Bliksem 1.10\n\n" ); 
   printstring( &outputfile,"\n"); 

   maketermends( 0, freetermspace ); 
   
   /* Default settings of flags:   */

   useres = 0;
   useparamod = 0;
   useeqrefl = 0;
   useeqfact = 0;
   usefactor = 0;

   usesimpsplitting = 0;
   usesimpdemod = 0;
   usesimpres = 0;

   resimpinuse = 2000;
   resimpclauses = 30000; 

   substype = SUBS_STANDARD; 
   backwardsubs = 1; 

   termapriori = ORD_NONE;
   nrlitorderings = 0; 
   
   maxweight =    15;
   maxdepth =     30000;
   maxlength =    MAXCLAUSELEN - 5;
   maxnrvars =    MAXNRVARS - 5; 
   maxselected =  MAXCLAUSES;
   maxclauses =   MAXCLAUSES; 

   selectoldest = 4;

   excuselevel = 0;     /* Excuse initial clauses only.    */ 
   increasemaxweight = 1; 

   showgenerated = 0;
   showkept = 0;
   showselected = 0;
   showdeleted = 0;
   showresimp = 1;
   showstatus = 2000;
   prologoutput = 0;
   nrgoals = 5000000;
   totalproof = 0;

   fullconversion = FULLCONV_STRUCTOS;

   weightsequal( );
   selectall( );
   ordersymbols( );

   /* Read the options. The options end either by End., or by Auto.  */

   i = 0;
   s = operatorstring( termspace [i] );
   while( s != (char*) NULL && strcmp( s, "End" ) != 0 && 
                               strcmp( s, "Auto" ) != 0 )
   {
      if( strcmp( s, "Set" ) == 0 )
      {
         if( tryassignment( i, "useres", &useres, 0, 1 ))
            goto nextoption;
         if( tryassignment( i, "useparamod", &useparamod, 0, 1 ))
            goto nextoption;

         if( tryassignment( i, "useeqrefl", &useeqrefl, 0, 1 ))
            goto nextoption;

         if( tryassignment( i, "useeqfact", &useeqfact, 0, 1 ))
            goto nextoption;

         if( tryassignment( i, "usefactor", &usefactor, 0, 1 ))
            goto nextoption;

         if( tryassignment( i, "usesimpsplitting", &usesimpsplitting, 
                            -1, 1000 ))
            goto nextoption;
         if( tryassignment( i, "usesimpdemod", &usesimpdemod, -1, 1000 ))
            goto nextoption;
         if( tryassignment( i, "usesimpres", &usesimpres, -1, 1000 ))
            goto nextoption;

         if( tryassignment( i, "resimpinuse", &resimpinuse, 10, 1000000 ))
            goto nextoption;
         if( tryassignment( i, "resimpclauses", &resimpclauses, 100, 9000000 ))
            goto nextoption;

         if( trysubstype( i ))
            goto nextoption;

         if( tryassignment( i, "selectoldest", &selectoldest, 0, 1000000 ))
            goto nextoption;

         if( trylitordering( i ))
            goto nextoption;
         if( trytermordering( i ))
            goto nextoption;
         if( trylitselect( i ))
            goto nextoption;

         if( tryassignment( i, "litapriori", &litapriori, 0, 1 ))
            goto nextoption;
         if( tryassignment( i, "termapriori", &termapriori, 0, 1 ))
            goto nextoption;
         if( tryassignment( i, "litaposteriori", &litaposteriori, 0, 1 ))
            goto nextoption;
         if( tryassignment( i, "termaposteriori", &termaposteriori, 0, 1 ))
            goto nextoption;
         if( tryassignment( i, "demodaposteriori", &demodaposteriori, 0, 1 ))
            goto nextoption;
         if( tryassignment( i, "ordereqreflfact", &ordereqreflfact, 0, 1 ))
            goto nextoption;

         if( tryassignment( i, "maxweight", &maxweight, -10, 10000 ))
            goto nextoption;
         if( tryassignment( i, "maxdepth", &maxdepth, -10, 10000 ))
            goto nextoption;
         if( tryassignment( i, "maxlength", &maxlength, -10, 10000 ))
            goto nextoption;
         if( tryassignment( i, "maxnrvars", &maxnrvars, -10, 10000 ))
            goto nextoption;

         if( tryassignment( i, "excuselevel", &excuselevel, -1, 1000000 ))
            goto nextoption;
         if( tryassignment( i, "increasemaxweight", &increasemaxweight, 
                            0, 100 ))
         {
            goto nextoption;
         }

         if( tryassignment( i, "maxselected", &maxselected, 0, 5000000 ))
            goto nextoption;
         if( tryassignment( i, "maxclauses", &maxclauses, 0, 5000000 ))
            goto nextoption; 

         if( tryassignment( i, "showgenerated", &showgenerated, 0, 1 ))
            goto nextoption;
         if( tryassignment( i, "showkept", &showkept, 0, 1 ))
            goto nextoption;
         if( tryassignment( i, "showselected", &showselected, 0, 1 ))
            goto nextoption;
         if( tryassignment( i, "showdeleted", &showdeleted, 0, 1 ))
            goto nextoption;
         if( tryassignment( i, "showresimp", &showresimp, 0, 1 ))
            goto nextoption;
         if( tryassignment( i, "showstatus", &showstatus, 100, 5000000 ))
            goto nextoption;

         if( tryassignment( i, "prologoutput", &prologoutput, 0, 4 ))
            goto nextoption;

         if( tryassignment( i, "nrgoals", &nrgoals, 0, 1000000 ))
            goto nextoption;

         if( tryassignment( i, "totalproof", &totalproof, 0, 1 ))
            goto nextoption;

         if( tryconversion( i, &fullconversion ))
            goto nextoption;
 
      }

      if( strcmp( s, "Set_order" ) == 0 )
      {
         operatorwoasp( operatororder, i, -1000000, 1000000 );
         goto nextoption;
      }
      if( strcmp( s, "Set_assoc" ) == 0 )  
      {
         operatorwoasp( operatorassoc, i, -1, 1 );
         goto nextoption;
      }
      if( strcmp( s, "Set_select" ) == 0 )
      {
         operatorwoasp( operatorselect, i, 0, 1 );
         goto nextoption;
      }
      if( strcmp( s, "Set_weight" ) == 0 )
      {
         operatorwoasp( operatorweights, i, -1000000, 1000000 );
         goto nextoption;
      }
      if( strcmp( s, "Set_noparamod" ) == 0 )
      {
         operatorwoasp( operatornoparamod, i, 0, 1 );
         goto nextoption;
      }

      printstring( &outputfile, "Did not recognize setting: ");
      printterm( &outputfile, i );
      errorwithoutline( "" );

   nextoption:
      i = termends [i];  
      s = operatorstring( termspace [i] ); 
   }
   if( s == (char*) NULL ) 
      errorwithoutline( "Settings should end with End or Auto" );

   i = termends [i];

   startofinput = i;
   endofinput = freetermspace; 

   if( strcmp( s, "Auto" ) == 0 )
   {
      convused = automode( );
      printjustconv( &outputfile, convused );
      goto derive;
   }

   if( strcmp( s, "End" ) != 0 )
      internalproblem( "End expected" );

usermode:
   nrtranslations = 0;   

   printtermlist( &outputfile, "Input:", startofinput, endofinput );
   printstring( &outputfile, "\n" ); 

   i = makeconversion( startofinput, endofinput, fullconversion );
   printjustconv( &outputfile, i );
   if( i != 0 ) internalproblem( "conversion not correct" );
   convused = 0;

   /* Complete the weights and the symbol order. Should be after */
   /* the conversion, because of the Skolem symbols.             */
   /* Sadly enough, the user cannot set the woasp of Skolem      */
   /* symbols, or structural names.                              */

   woaspskolem( ); 
   
derive:
   freetermspace2 = freetermspace;
   freeclauses2 = freeclauses;

   printoptions( &outputfile, convused );

restart:
   printstring( &outputfile, "\nStarting Search:\n\n" ); 

   i = mainloop( convused, 0 );

   if( ! i && increasemaxweight )
   {
      freetermspace = freetermspace2;
      freeclauses = freeclauses2;   

      maxweight += increasemaxweight ; 
#if 0
      maxdepth = maxweight / 3;   /* However this is often better.          */
#endif
      printstring( &outputfile, "Increased maxclauses to ");
      printint( &outputfile, maxclauses );
      printstring( &outputfile, "\n" );

      goto restart;
   }
   printstatistics( &outputfile ); 
   return 0;
}





