


/****************************************************************/
/*    conversion.c contains the functions for the conversions   */
/*    from FOL into clauses.                                    */
/*    Bliksem 1.10                                              */
/****************************************************************/


#define CLASS extern
#include "basics.h"
#include "resolution.h"
#include "extfunc.h"


/***********************************************************/
/* Variables in the context of which a subformula occurs.  */
/***********************************************************/

static struct 
{
   int var;
   int marked;
}
   skolvars [ MAXNRVARS ];
static int nrskolvars;


static void printskolvars( )
{
   int i;

   printstring( &outputfile, "[ " );
   for( i = 0; i < nrskolvars; i ++ )
   {
      printoperator( &outputfile, skolvars [i]. var );
      if( i < nrskolvars - 1 ) 
         printstring( &outputfile, ", " );
   }
   printstring( &outputfile, " ]\n" );
}


/* Collect the bound variables in the context of which the term on  */
/* position occurs. They are stored as operators. In the case that  */
/* an operator occurs twice, the last occurrence is kept.           */

static void collectskolvars( termbase, position )
int termbase;
int position;
{
   int i1, i2, j;
   if( position < termbase || position >= termends [ termbase ] )
      exit(0);

   nrskolvars = 0;
   while( termbase != position )
   {
      if( termspace [ termbase ] == OP_FORALL || 
          termspace [ termbase ] == OP_EXISTS )
      {
         if( nrskolvars >= MAXNRVARS - 3 ) 
            errorwithoutline( "too many variables in Skolem term" ); 
         skolvars [ nrskolvars ++ ]. var = termspace [ termbase + 1 ];
      }

      /* find subterm which contains position.  */

      termbase ++;
      while( termends [ termbase ] <= position )
         termbase = termends [ termbase ];
   }

   /* In the case that a variable is bound more than one, keep the  */
   /* last.                                                         */

   i1 = 0;
   i2 = 0;
   while( i1 < nrskolvars )
   {
      for( j = i1 + 1; j < nrskolvars; j ++ )
      {
         if( skolvars [i1]. var == skolvars [j]. var )
            goto dontkeep;
      }
      skolvars [ i2 ++ ]. var = skolvars [i1]. var;
   dontkeep:
      i1 ++ ;
   }
   nrskolvars = i2; 
}


/* delete the variables that are not free from skolvars [].   */

static void deletenotfree( term )
int term;
{
   int i1, i2;

   i1 = 0; 
   i2 = 0;
   while( i1 < nrskolvars )
   {
      if( fovarfree( skolvars [ i1 ]. var, term ))
      {
         skolvars [ i2 ++ ]. var = skolvars [ i1 ]. var;
      }
      i1 ++ ;
   }
   nrskolvars = i2;
}


/* Make a Skolem term of type flag. */

static int makeskolemterm( operator )
int operator;
{
   int res, i;

   if( arities [ operator ] != nrskolvars ) 
      exit(0);

   res = freetermspace;
   if( freetermspace + nrskolvars >= MAXTERMSPACE - 10 ) 
      errorwithoutline( "termspace is full" );

   termspace [ freetermspace ] = operator;
   for( i = 0; i < nrskolvars; i ++ )
      termspace [ freetermspace + 1 + i ] = skolvars [i]. var;
   freetermspace += ( 1 + nrskolvars ); 
   maketermends( res, freetermspace );
   return res;
}



/********************************************************************/
/* Each conversion sequence (called fullconversion, fullconv etc),  */
/* which is selectable by the user or the automode, consists of     */
/* a sequence of basic steps.                                       */
/* The function expandfullconv( ) produces the sequence of          */
/* basic steps from the fullconversion.                             */
/* The conversion sequences are called fullconversions, the         */
/* basic steps are called just conversions.                         */
/********************************************************************/

 
static int fullconv_none [] = { -1 };
static int fullconv_standard [] =
   { CONV_SIMPQUANTFACT, 
     CONV_STRUCTNA2, CONV_NNF, 
     CONV_SIMPANTIPRENEX, CONV_SKOLEMNG,
     CONV_STRUCTFACTOS2, CONV_NNF, CONV_FACTOR, -1 };
static int fullconv_struct [] =
   { CONV_SIMPQUANTFACT, 
     CONV_STRUCTNA1, CONV_NNF, 
     CONV_STRUCTFACT1, CONV_NNF, CONV_SKOLEMNG,
     CONV_FACTOR, -1 };
static int fullconv_structos [] =
   { CONV_SIMPQUANTFACT,
     CONV_STRUCTNA2, CONV_NNF, 
     CONV_STRUCTFACTOS1, CONV_NNF, CONV_SKOLEMNG, 
     CONV_FACTOR, -1 };
static int fullconv_guarded [] =
   { CONV_NNF, CONV_STRUCTGUARD, 
     CONV_NNF, CONV_SKOLEM, 
     CONV_STRUCTFACTOS2, CONV_NNF, 
     CONV_FACTOR, -1 };
static int fullconv_2var [] =
   { CONV_NNF, CONV_SKOLEM, CONV_FACTOR, -1 };


static int *expandfullconv( fullconversion )
int fullconversion;
{
   switch( fullconversion )
   {
   case FULLCONV_NONE:
      return fullconv_none;
   case FULLCONV_STANDARD:
      return fullconv_standard;
   case FULLCONV_GUARDED:
      return fullconv_guarded;
   case FULLCONV_STRUCT:
      return fullconv_struct;
   case FULLCONV_STRUCTOS:
      return fullconv_structos;
   case FULLCONV_2VAR:
      return fullconv_2var;
   default:
      errorwithoutline( 
         "Unknown full conversion. This is an internal problem" );
   }
}



/* Print a full conversion.     */

void printfullconversion( f, fullconv )
bliksemfile *f;
int fullconv;
{
   switch( fullconv )
   {
   case FULLCONV_NONE:
      printstring( f, "none" );
      return;
   case FULLCONV_STANDARD:
      printstring( f, "standard" );
      return;
   case FULLCONV_GUARDED:
      printstring( f, "guarded" );
      return;
   case FULLCONV_STRUCT:
      printstring( f, "structural" );
      return;
   case FULLCONV_STRUCTOS:
      printstring( f, "one sided structural" );
      return;
   case FULLCONV_2VAR:
      printstring( f, "Two variable class" );
      return;
   default:
      exit(1);
   }
}



/****************************************************************/
/* subterms is used for the decomposition of and's and or's.    */
/****************************************************************/


#define MAXSUBTERMS  100 
static struct
{
   int subterm;
}
   subterms [ MAXSUBTERMS ];
static int nrsubterms;


static void printsubterms( )
{
   int i;

   printstring( &outputfile, "Subterms:\n\n");
   for( i = 0; i < nrsubterms; i ++ )
   {
      printterm( &outputfile, subterms [i]. subterm );
      printstring( &outputfile, "\n" );
   }
   printstring( &outputfile, "\n" );
}


/* collect the subterms of term, as long as they are built by operator. */

static void collectsubterms( operator, term )
int operator;
int term;
{
   int i;

   nrsubterms = 0;
   
   i = term;
   while( i < termends [ term ] )
   {
      if( termspace [i] != operator || nrsubterms >= MAXSUBTERMS - 5 ) 
      {
         subterms [ nrsubterms ++ ]. subterm = i;
         i = termends [i];
      }
      else
         i ++ ;
   }    
}

/* Build a term from subterms [ t1 .. t2 - 1 ]. subterm using  */
/* operator.                                                   */

static int buildfromsubterms( t1, t2, operator )
int t1, t2;
int operator;
{
   int i;
   int res;

   res = freetermspace; 

   if( t1 >= t2 ) exit(0);

   if( freetermspace + ( t2 - t1 ) >= MAXTERMSPACE - 10 ) 
      errorwithoutline( "termspace is full" );

   for( i = t1 + 1; i < t2; i ++ )
      termspace [ freetermspace ++ ] = operator;
   
   for( i = t1; i < t2; i ++ )
      (void) copyterm( subterms [i]. subterm );
   
   maketermends( res, freetermspace );
   return res;
}


/* Produce a permutation of subterms []. subterm, and an n, such that  */
/* op is not free in subterms [ 0 .. n - 1 ]. subterm, and is free in  */
/* subterms [ n .. nrsubterms - 1 ]. subterm.                          */

static int sortbyfree( op )
int op;
{
   int i1, i2;
   int c;

   i1 = 0;
   i2 = nrsubterms;

   while( i1 < i2 )
   {
      if( fovarfree( op, subterms [ i1 ]. subterm ))
      {
         i2 -- ;
         c = subterms [ i1 ]. subterm;
         subterms [ i1 ]. subterm = subterms [ i2 ]. subterm;
         subterms [ i2 ]. subterm = c;
      }
      else
      {
         i1 ++ ;
      }
   }
   return i1;
}




/******************************************************************/
/* printjustconv( f, juststart, justend ) prints a sequence of    */
/* justifications made by a conversion.                           */
/******************************************************************/


void printjustconv( f, nr )
bliksemfile *f;
int nr;
{
   int i;
   int conv;
   int res;

   if( totalproof == 0 ) return;

   if( prologoutput )
   {
      printstring( &outputfile, "% 0123456789" );
      printstring( &outputfile, "\n\n" );
   }
   else
   {
      printstring( f, "\n\n");
      printstring( f, "Conversion number ");
      printint( f, nr + 1 );
      printstring( f, ":\n\n");   
   }

   i = translations [ nr ]. startofjust;
   while( i < translations [ nr ]. endofjust )
   {
      conv = (justifications [i] >> 8);
      switch( conv )
      {
      case CONV_NNF:
         printstring( f, "nnf" );
         goto simple;
      case CONV_SIMPQUANTFACT:
         printstring( f, "quantfactor" );
         goto simple;
      case CONV_FACTOR:
         printstring( f, "factor" );
         goto simple;
      simple:
         printstring( f, "( " );
         printint( f, justifications [i] & 0X0FF );
         printstring( f, ", ");
         printint( f, justifications [ i + 1 ] );
         printstring( f, " ).\n");
         i += 2;
         goto next;
      case CONV_SKOLEM:
      case CONV_SKOLEMFREE:
      case CONV_SKOLEMNG:
         printstring( f, "skolem" );
         goto notsimple;
      case CONV_STRUCTNA1:
      case CONV_STRUCTNA2:
      case CONV_STRUCTFACT1:
      case CONV_STRUCTFACT2:
      case CONV_STRUCTFACTOS1:
      case CONV_STRUCTFACTOS2:
      case CONV_STRUCTGUARD:
         printstring( f, "struct" );
         goto notsimple;
      case CONV_SIMPANTIPRENEX:
         printstring( f, "antiprenex" );
         goto notsimple;
      notsimple:
         printstring( f, "( " );

         res = freetermspace;
         i ++ ;
         while( justifications [i] >= 0 )
         {
            termspace [ freetermspace ++ ] = justifications [ i ++ ];
            if( freetermspace >= MAXTERMSPACE - 10 ) 
               errorwithoutline( "termspace is full" );
         }
         maketermends( res, freetermspace );
         printterm( f, res );
         freetermspace = res;
         i ++ ;
         printstring( f, ", " );
         printint( f, justifications [ i ++ ] );
         printstring( f, " ).\n" );
         goto next;
      default:
         exit(0);
      }
   next:
      ; 
   }

   if( !prologoutput )
   {
      printtermlist( f, "Result: ", translations [ nr ]. startofresult, 
                                    translations [ nr ]. endofresult );
   }

   if( prologoutput )
   {
      printstring( &outputfile, "\n" );
      printstring( &outputfile, "% 0123456789\n\n" );
   }
   else
   {
      printstring( &outputfile, "\n\n" );
   } 
}



/* store a complicated (Skolem/struct/antiprenex) replacement. */

static void storecomplrepl( was, conversion, resultbase, becomes )
int was;
int conversion;
int resultbase;
int becomes;
{
   int i;

   if( totalproof )
   {
      justifications [ freejustifications ++ ] = ( conversion << 8 ); 
      
      for( i = 0; i < nrskolvars; i ++ )
      {
         justifications [ freejustifications ++ ] = OP_FORALL;
         justifications [ freejustifications ++ ] = skolvars [i]. var;
         if( freejustifications >= MAXJUSTIFICATIONS - 10 ) 
            errorwithoutline( "justifications is full" );
      }
      justifications [ freejustifications ++ ] = OP_EQUIV;
      for( i = was; i < termends [ was ]; i ++ )
      {
         justifications [ freejustifications ++ ] = termspace [i];
         if( freejustifications >= MAXJUSTIFICATIONS - 10 ) 
            errorwithoutline( "justifications is full" );
      }
      for( i = becomes; i < termends [ becomes ]; i ++ )
      {
         justifications [ freejustifications ++ ] = termspace [i];
         if( freejustifications >= MAXJUSTIFICATIONS - 10 ) 
            errorwithoutline( "justifications is full" );
      }
      justifications [ freejustifications ++ ] = -1;
      justifications [ freejustifications ++ ] = becomes - resultbase;
   }   
}


/* store a simple (rewrite) replacement.    */

static void storesimprepl( conversion, rulenr, resultbase, result )
int conversion;
int rulenr;
int resultbase;
int result;
{
   if( totalproof )
   {
      if( freejustifications >= MAXJUSTIFICATIONS - 10 ) 
         errorwithoutline( "justifications is full" );

      justifications [ freejustifications ] = ( conversion << 8 ) | rulenr;
      justifications [ freejustifications + 1 ] = result - resultbase; 
      freejustifications += 2;
   }
}



/*****************************************************************/
/*     Skolem tranformations.                                    */
/*****************************************************************/


static int standardskolem( resultbase, formula, conversion )
int resultbase;
int formula;
int conversion;
{
   int skolterm, subst;
   int existential, existsbody;

   existential = formula;
   while( existential < termends [ formula ] ) 
   {
      if( termspace [ existential ] == OP_EXISTS ) 
      {
         collectskolvars( formula, existential );
         if( conversion == CONV_SKOLEMFREE )
            deletenotfree( existential ); 
     
         existsbody = copyterm( existential + 2 );  
         skolterm = makeskolemterm( 
                     makefreshoperator( FLAG_SKOLEM, nrskolvars ));
         subst = fosubst( existsbody, termspace [ existential + 1 ], 
                          skolterm );
         deletetermspace( existsbody, subst - existsbody );
         maketermends( existsbody, freetermspace );

         storecomplrepl( existential, conversion, resultbase, existsbody );
         existential = termends [ existential ];

         /* Copy the rest of the formula:      */

         while( existential < termends [ formula ] )
         {
            termspace [ freetermspace ++ ] = termspace [ existential ++ ];
            if( freetermspace >= MAXTERMSPACE - 10 ) 
               errorwithoutline( "termspace is full" );
         }
         return 1;
      }
      termspace [ freetermspace ++ ] = termspace [ existential ++ ];
      if( freetermspace >= MAXTERMSPACE - 10 ) 
         errorwithoutline( "termspace is full" );
   }   
   return 0;
}


/* mark the free variables of term that occur in skolvars [].   */

static void markfreevars( term )
int term;
{
   int i;

   for( i = 0; i < nrskolvars; i ++ )
   {
      if( skolvars [i]. marked == 0 &&
          fovarfree( skolvars [i]. var, term ))
         skolvars [i]. marked = 1;
   }
}


/* Sort the variables in skolvars [] by their free variables that   */
/* occur in skolvars []. We use a greedy algorithm.                 */

static void sortsubterms( )
{
   int sort;
   int i, j;
   int bestsubterm;
   int cnt, bestcnt;
   int first, bestfirst;
   int c;

   for( i = 0; i < nrskolvars; i ++ )
      skolvars [i]. marked = 0;

   for( sort = 0; sort < nrsubterms - 1; sort ++ )
   {
      bestcnt = MAXNRVARS + MAXNRVARS;
      bestfirst = 0;

      bestsubterm = 0;
      for( i = sort; i < nrsubterms; i ++ )
      {
         /* Find subterm that has smallest number of unmarked free */
         /* variables, and put it on position sort.                */

         cnt = 0;
         first = MAXNRVARS + MAXNRVARS;
         for( j = 0; j < nrskolvars; j ++ )
         {
            if( skolvars [j]. marked == 0 && 
                fovarfree( skolvars [j]. var, subterms [i]. subterm ))
            {
               cnt ++ ;
               if( j < first )
                  first = j;
            }
         }
         if( (cnt < bestcnt) || 
             (cnt == bestcnt && first > bestfirst ))
         {
            bestsubterm = i;
            bestcnt = cnt;
            bestfirst = first;
         }
      }
   
      /* Now we know the best. */
      /* Permute it.           */
 
      c = subterms [ bestsubterm ]. subterm;
      subterms [ bestsubterm ]. subterm = subterms [ sort ]. subterm;
      subterms [ sort ]. subterm = c;
 
      /* Mark its free variables. */

      markfreevars( subterms [ sort ]. subterm );
   }
}


/* Skolemize using Nonnengart's algorithm.   */

static int nonnengartskolem( resultbase, formula )
int resultbase;
int formula;
{
   int totalresult;
   int operator;
   int skolterm, subst;
   int existential, existsbody;
   int i, st;
 
   existential = formula;
   while( existential < termends [ formula ] )
   {
      if( termspace [ existential ] == OP_EXISTS ) 
      {
         collectskolvars( formula, existential );
         deletenotfree( existential );

         collectsubterms( OP_AND, existential + 2 );
         sortsubterms( );

         totalresult = freetermspace;

         if( freetermspace + nrsubterms >= MAXTERMSPACE - 10 ) 
            errorwithoutline( "termspace is full" );

         for( i = 0; i < nrsubterms - 1; i ++ )
            termspace [ freetermspace ++ ] = OP_AND;
         for( i = 0; i < nrskolvars; i ++ )
            skolvars [i]. marked = 0;

         operator = makefreshoperator( FLAG_SKOLEM, nrskolvars );

         for( st = 0; st < nrsubterms; st ++ )
         {
            markfreevars( subterms [ st ]. subterm );
            for( i = 0; i < nrskolvars; i ++ )
               if( skolvars [i]. marked == 0 )
               {
                  termspace [ freetermspace ++ ] = OP_FORALL;
                  termspace [ freetermspace ++ ] = skolvars [i]. var;
                  if( freetermspace >= MAXTERMSPACE - 10 ) 
                     errorwithoutline( "termspace is full" );
               }
            existsbody = copyterm( subterms [ st ]. subterm );
            skolterm = makeskolemterm( operator );
            subst = fosubst( existsbody, termspace [ existential + 1 ], 
                             skolterm ); 
            deletetermspace( existsbody, subst - existsbody );
         }
         maketermends( totalresult, freetermspace );
         storecomplrepl( existential, CONV_SKOLEMNG, resultbase, totalresult );
 
         existential = termends [ existential ];
         while( existential < termends [ formula ] )
         {
            termspace [ freetermspace ++ ] = termspace [ existential ++ ] ;
            if( freetermspace >= MAXTERMSPACE - 10 ) 
               errorwithoutline( "termspace is full" );
         }
         return 1;
      }
      termspace [ freetermspace ++ ] = termspace [ existential ++ ];
      if( freetermspace >= MAXTERMSPACE - 10 ) 
         errorwithoutline( "termspace is full" );

   }
   return 0;
}
 

/* Construct Skolemization.    */

static int convskolem( startofinput, endofinput, conversion )
int startofinput;
int endofinput;
int conversion;
{
   int formula;
   int rewr;
   int resultbase;

   resultbase = freetermspace;
   rewr = 0;
   formula = startofinput;
   while( formula < endofinput )
   {
      switch( conversion )
      {
      case CONV_SKOLEMFREE:
      case CONV_SKOLEM:
         if( standardskolem( resultbase, formula, conversion ))
            rewr = 1;
         break; 
      case CONV_SKOLEMNG:
         if( nonnengartskolem( resultbase, formula ))
            rewr = 1;
         break;
      default:
         exit(0);
      }
      formula = termends [ formula ];
   }   
   return rewr;
}

 
/**************************************************************/
/* antiprenexing tries to move formulae outside the scope of  */
/* an exists, that do not contain the quantified variable.    */
/* <y> a(y) & b ==> b & <y> a(y).                             */
/**************************************************************/


static int antiprenex( resultbase, formula )
int resultbase;
int formula;
{
   int existential;
   int firstfree;
   int i;
   int res;

   existential = formula;
   while( existential < termends [ formula ] )
   {
      if( termspace [ existential ] == OP_EXISTS ) 
      {
         /* This is used only in the report.   */
         collectskolvars( formula, existential );
         deletenotfree( existential );

         collectsubterms( OP_AND, existential + 2 );
         firstfree = sortbyfree( termspace [ existential + 1 ] );
         if( firstfree == 0 ) goto nosig;
         if( firstfree == nrsubterms )
         {
            res = freetermspace;
            i = existential + 2;
            while( i < termends [ formula ] )
            {
               termspace [ freetermspace ++ ] = termspace [ i ++ ];
               if( freetermspace >= MAXTERMSPACE - 10 ) 
                  errorwithoutline( "termspace is full" );
            }
            maketermends( res, freetermspace );
            storecomplrepl( existential, CONV_SIMPANTIPRENEX, 
                            resultbase, res );
            return 1;
         }

         res = freetermspace;
         termspace [ freetermspace ++ ] = OP_AND;
         buildfromsubterms( 0, firstfree, OP_AND );
         termspace [ freetermspace ++ ] = OP_EXISTS;
         termspace [ freetermspace ++ ] = termspace [ existential + 1 ];
         buildfromsubterms( firstfree, nrsubterms, OP_AND );

         maketermends( res, freetermspace );         
         storecomplrepl( existential, CONV_SIMPANTIPRENEX, 
                         resultbase, res );

         existential = termends [ existential ]; 
         while( existential < termends [ formula ] )
         {
            termspace [ freetermspace ++ ] = termspace [ existential ++ ];
            if( freetermspace >= MAXTERMSPACE - 10 ) 
               errorwithoutline( "termspace is full" );
         }
         return 1;
      }
   nosig:
      if( freetermspace >= MAXTERMSPACE - 10 ) 
         errorwithoutline( "termspace is full" );
      termspace [ freetermspace ++ ] = termspace [ existential ++ ] ;
   }
   return 0;
}
              


static int convantiprenex( startofinput, endofinput, conversion )
int startofinput;
int endofinput;
int conversion; 
{
   int formula;
   int rewr;
   int resultbase;

   resultbase = freetermspace; 
   rewr = 0;
   for( formula = startofinput; formula < endofinput;
        formula = termends [ formula ] )
   {
      if( antiprenex( resultbase, formula ))
         rewr = 1;
   }
   return rewr;
}



/******************************************************************/
/* Structural replacements replace subformulae by a fresh atom +  */
/* a definition of the fresh atom.                                */
/******************************************************************/

/* the number of atoms after <->' are completely expanded.   */


static double equivcomp( formula, subform )
int formula;
int subform;
{
   int op;

   if( formula == subform ) 
      return 1.0;

   op = termspace [ formula ++ ];
   switch( op )
   {
   case OP_FALSE:
   case OP_TRUE:
      return 0.0;
   case OP_NOT:
      return equivcomp( formula, subform );
   case OP_FORALL:
   case OP_EXISTS:
      return equivcomp( formula + 1, subform );
   case OP_OR:
   case OP_AND:
   case OP_IMPLIES:
      return equivcomp( formula, subform ) + 
             equivcomp( termends [ formula ], subform );
   case OP_EQUIV:
      return 2.0 * ( equivcomp( formula, subform ) +
                     equivcomp( termends [ formula ], subform ));
   default:
      return 1.0;
   }
}


/* the number of atoms in the formula.   */

static double formcomp( formula, subform )
int formula;
int subform;
{
   int op;
 
   if( formula == subform )
      return 1.0;
 
   op = termspace [ formula ++ ];
   switch( op )
   {
   case OP_FALSE:
   case OP_TRUE:
      return 0.0;
   case OP_NOT:
      return formcomp( formula, subform );
   case OP_FORALL:
   case OP_EXISTS:
      return formcomp( formula + 1, subform );
   case OP_OR:
   case OP_AND:
   case OP_IMPLIES:
   case OP_EQUIV:
      return formcomp( formula, subform ) +
             formcomp( termends [ formula ], subform );
   default:
      return 1.0;
   }
}
 

static int replacable( subform )
int subform;
{
   int op;

   op = termspace [ subform ];
   if( op == OP_NOT || op == OP_AND || op == OP_OR || op == OP_IMPLIES || 
       op == OP_EQUIV || op == OP_FORALL || op == OP_EXISTS )
      return 1;
   return 0;
}



static int repl_na( formula, tolerance )
int formula;
double tolerance;
{
   int subform, bestsubform;
   int bestfreevars;

   double ec;

   ec = equivcomp( formula, -1 ) - formcomp( formula, -1 );
   if( ec <= tolerance ) return -1;

   bestfreevars = MAXNRVARS + MAXNRVARS + MAXNRVARS;
   bestsubform = -1;

   for( subform = formula; subform < termends [ formula ]; subform ++ )
   {
      if( replacable( subform ))
      {
         if( (1.0 + equivcomp( subform, -1 )) < ec &&
             equivcomp( formula, subform ) - 
             formcomp( formula, subform ) < ec )
         {
            collectskolvars( formula, subform );
            deletenotfree( subform );
            if( nrskolvars < bestfreevars )
            {
               bestfreevars = nrskolvars;
               bestsubform = subform;
            }
         }
      }
   }
   return bestsubform;
}


/* polarity is 1 or -1.                                               */
/* the number of disjuncts if formula were to be factored into CNF.   */


static double factcomp( formula, subform, polarity )
int formula;
int subform;
{
   int op;

   if( polarity != -1 && polarity != 1 ) exit(0);
 
   if( formula == subform )
      return 1.0;
 
   op = termspace [ formula ];
   formula ++ ;

   switch( op )
   {
   case OP_FALSE:
   case OP_TRUE:
      return 1.0;
   case OP_NOT:
      return factcomp( formula, subform, -polarity );
   case OP_IMPLIES:
      if( polarity == 1 )
         return factcomp( formula, subform, -1 ) *
                factcomp( termends [ formula ], subform, 1 );
      else
         return factcomp( formula, subform, 1 ) +
                factcomp( termends [ formula ], subform, -1 );
   case OP_EQUIV:
      if( polarity == 1 )
         return factcomp( formula, subform, -1 ) *
                factcomp( termends [ formula ], subform, 1 ) +
                factcomp( formula, subform, 1 ) *
                factcomp( termends [ formula ], subform, -1 );      
      else
         return (factcomp( formula, subform, 1 ) +
                 factcomp( termends [ formula ], subform, -1 )) *
                (factcomp( formula, subform, -1 ) + 
                 factcomp( termends [ formula ], subform, 1 ));
   case OP_FORALL:
   case OP_EXISTS:
      return factcomp( formula + 1, subform, polarity );
   case OP_OR:
      if( polarity == 1 )
         return factcomp( formula, subform, 1 ) *
                factcomp( termends [ formula ], subform, 1 );
      else
         return factcomp( formula, subform, -1 ) +
                factcomp( termends [ formula ], subform, -1 );
   case OP_AND:
      if( polarity == 1 )
         return factcomp( formula, subform, 1 ) +
                factcomp( termends [ formula ], subform, 1 );
      else
         return factcomp( formula, subform, -1 ) * 
                factcomp( termends [ formula ], subform, -1 );
   default:
      return 1.0;
   }
}


/* The number of ands that formula has on top level.   */
 
static double andsontop( formula, subform, polarity )
int formula;
int subform;
int polarity;
{
   int op;

   if( polarity != -1 && polarity != 1 )
      exit(0);

   if( formula == subform ) 
      return 1.0;

   op = termspace [ formula ++ ];
   switch( op )
   {
   case OP_FALSE:
   case OP_TRUE:
      return 1.0;
   case OP_IMPLIES:
      if( polarity == 1 )
         return 1.0;
      else
         return andsontop( formula, subform, 1 ) +
                andsontop( termends [ formula ], subform, -1 );
   case OP_OR:
      if( polarity == 1 )
         return 1.0;
      else
         return andsontop( formula, subform, -1 ) + 
                andsontop( termends [ formula ], subform, -1 );
   case OP_AND:
      if( polarity == 1 )
         return andsontop( formula, subform, 1 ) +
                andsontop( termends [ formula ], subform, 1 ); 
      else
         return 1.0;
   case OP_EQUIV:
      if( polarity == 1 )
         return 2.0;
      else
         return 1.0;
   case OP_EXISTS:
   case OP_FORALL:
      return andsontop( formula + 1, subform, polarity );
   case OP_NOT:
      return andsontop( formula, subform, -polarity );
   default:
      return 1.0;
   }
}
   


static int repl_fact( formula, tolerance, defoperator )
int formula;
double tolerance;
int defoperator;
{
   int subform, bestsubform;
   int bestfreevars;
 
   double ec;

   ec = factcomp( formula, -1, 1 ) - andsontop( formula, -1, 1 );
   if( ec <= tolerance ) return -1;
 
   bestfreevars = MAXNRVARS + MAXNRVARS + MAXNRVARS;
   bestsubform = -1;

   for( subform = formula; subform < termends [ formula ]; subform ++ )
   {
      if( replacable( subform ))
      {
         if( factcomp( formula, subform, 1 ) - 
             andsontop( formula, subform, 1 ) < ec )
         {
             if( defoperator == OP_IMPLIES && 
                    factcomp( subform, -1, 1 ) - 1.0 < ec ||
                 defoperator == OP_EQUIV &&  
                    factcomp( subform, -1, -1 ) +  
                    factcomp( subform, -1, 1 ) - 2.0 < ec )
            {
               collectskolvars( formula, subform );
               deletenotfree( subform );
               if( nrskolvars < bestfreevars )
               {
                  bestfreevars = nrskolvars;
                  bestsubform = subform;
               }
            }
         }
      }
   }
   return bestsubform;
}
 
 

static int repl_guarded( formula )
int formula;
{
   int formend;

   formend = termends [ formula ];
   while( termspace [ formula ] == OP_FORALL ) 
      formula += 2;

   if( termspace [ formula ] == OP_IMPLIES ) 
   {
      formula ++ ;
      formula = termends [ formula ];
      while( termspace [ formula ] == OP_FORALL ) 
         formula += 2;
   }

   while( formula < formend )
   {
      if( termspace [ formula ] == OP_FORALL ) 
         return formula;
      formula ++ ;
   }
   return -1;
}


/* Make one pass of structural transformations.        */
/* Return 1 if a replacement was made.                 */

static int convstruct( startofinput, endofinput, conversion )
int startofinput;
int endofinput;
int conversion;
{
   int rewr;
   int formula;
   int subterm, replaced, skolterm;
   int op, i;
   int defoperator; 
   int resultbase;

   resultbase = freetermspace;
 
   rewr = 0;
   for( formula = startofinput; formula < endofinput; 
        formula = termends [ formula ] )
   {
      switch( conversion )
      {
         /* make replaced equal to the subterm of formula that should  */
         /* be replaced, or -1 if no replacement can be made.          */

      case CONV_STRUCTNA1:
         replaced = repl_na( formula, 2.0 );
         defoperator = OP_EQUIV;
         break;
      case CONV_STRUCTNA2:
         replaced = repl_na( formula, (double) MAXEQUIVDIFF );
         defoperator = OP_EQUIV;
         break;
      case CONV_STRUCTFACT1:
         replaced = repl_fact( formula, 2.0, OP_EQUIV );
         defoperator = OP_EQUIV;
         break;
      case CONV_STRUCTFACT2:
         replaced = repl_fact( formula, (double) MAXANDINOR, OP_EQUIV );
         defoperator = OP_EQUIV;
         break;
      case CONV_STRUCTFACTOS1:
         replaced = repl_fact( formula, 2.0, OP_IMPLIES );
         defoperator = OP_IMPLIES;
         break;
      case CONV_STRUCTFACTOS2:
         replaced = repl_fact( formula, (double) MAXANDINOR, OP_IMPLIES );
         defoperator = OP_IMPLIES;
         break;
      case CONV_STRUCTGUARD:
         replaced = repl_guarded( formula );
         defoperator = OP_IMPLIES;
         break;
      default:
         errorwithoutline( "unknown conversion" );
      }

      subterm = formula;
      while( subterm < termends [ formula ] )
      {
         if( subterm == replaced )
         {
            collectskolvars( formula, replaced );
            deletenotfree( replaced );

            op = makefreshoperator( FLAG_SUBFORM, nrskolvars );
            (void) makeskolemterm( op );

            /* copy the rest.      */

            subterm = termends [ subterm ];
        
            if( freetermspace + ( termends [ formula ] - subterm )  
                >= MAXTERMSPACE - 10 ) 
               errorwithoutline( "termspace is full" );

            while( subterm < termends [ formula ] )
               termspace [ freetermspace ++ ] = termspace [ subterm ++ ] ;

            /* Produce the defining axiom:   */

            if( freetermspace + 2 * nrskolvars >= MAXTERMSPACE - 10 ) 
               errorwithoutline( "termspace is full" );

            for( i = 0; i < nrskolvars; i ++ )
            {
               termspace [ freetermspace ++ ] = OP_FORALL;
               termspace [ freetermspace ++ ] = skolvars [i]. var;
            }
           
            termspace [ freetermspace ++ ] = defoperator; 

            skolterm = makeskolemterm( op );
            (void) copyterm( replaced );

            storecomplrepl( replaced, conversion, resultbase, skolterm );

            rewr = 1;
            goto replacement;
         }
         termspace [ freetermspace ++ ] = termspace [ subterm ++ ];
         if( freetermspace >= MAXTERMSPACE - 10 ) 
            errorwithoutline( "termspace is full" );
      }
   replacement:
      ;
   }
   return rewr;
}
 
/* This function should produce the complete clauses in the future.   */

static int convfactor( startofinput, endofinput, conversion )
int startofinput, endofinput;
int conversion;
{
   int formula;
   int rewr;
   int resultbase;

   resultbase = freetermspace;
   rewr = 0;
   for( formula = startofinput; formula < endofinput;
        formula = termends [ formula ] )
   {
      if( termspace [ formula ] == OP_AND ) 
      {
         storesimprepl( conversion, 4, resultbase, freetermspace );

         formula ++ ;
         (void) copyterm( formula );
         formula = termends [ formula ];
         (void) copyterm( formula );
         rewr = 1;
      }
      else
         (void) copyterm( formula );
   }
   return rewr;
}



/* Try to apply a simple rewrite rule on term.        */
/* term is a reference into termspace. lhs and rhs    */
/* are pointers to sequences of integers representing */
/* the rewrite rule in the standard manner, but not   */
/* in termspace. Small negative integers are used     */
/* for variables.                                     */
/* If the rule can be applied then the result is      */
/* created on termspace [ freetermspace ] and the     */
/* position of the result is returned. Otherwise      */
/* -1 is returned.                                    */

#define MAXRULESIZE 10 
 
static int applyrewriterule( term, lhs, rhs, usecondition, condition )
int term;
int *lhs; 
int *rhs;
int usecondition;
int condition( int, int, int, int );
{
   int i, j;
   int f, remain;

   int vars [ MAXNRVARS ];
   int vals [ MAXNRVARS ]; 
   int sortedvals [6];

   int nrassignments;

   nrassignments = 0;
   i = term; 
   remain = 1;
   while( remain > 0 ) 
   {
      f = *lhs; 
      if( f >= 0 )
      {
         if( f != termspace [i] )
            return 0;
         remain += arities [f] - 1;
         lhs ++ ;
         i ++ ; 
      }
      else
      {
         if( nrassignments >= MAXNRVARS - 2 )  
            errorwithoutline( "increase MAXNRVARS" ); 
         vars [ nrassignments ] = f;
         vals [ nrassignments ] = i;
         nrassignments ++ ;
         remain -- ;
         lhs ++ ;
         i = termends [i];  
      }
   }

   /* We produced an assignment. Check its consistency */

   for( i = 0; i < nrassignments; i ++ )
      for( j = 0; j < i; j ++ )
         if( vars [i] == vars [j] )
         {
            if( ! termequal( vals [i], vals [j] ))
               return 0;
         }

   /* If condition( ) is different from NULL, we have to call condition. */

   if( usecondition )
   {
      if( nrassignments >= 4 ) 
         exit(0);

      sortedvals [0] = -1;
      sortedvals [1] = -1;
      sortedvals [2] = -1;
      sortedvals [3] = -1;

      for( i = 0; i < 4; i ++ )
      {
         for( j = 0; j < nrassignments; j ++ )
         {
            if( vars [j] == -i - 1 )
            {
               sortedvals [i] = vals [j];
               goto found;
            }
         }
      found:
         ; 
      }
  
      if( !condition( sortedvals [0], sortedvals [1], 
                      sortedvals [2], sortedvals [3] ))
         return 0;
   }

   /* The assignment is consistent, and has passed condition( ).    */
   /* We can make the rewrite.                                      */

   remain = 1;
   while( remain > 0 ) 
   {
      if( freetermspace >= MAXTERMSPACE - 5 ) 
         errorwithoutline( "termspace is full" ); 

      f = *rhs;
      if( f >= 0 )
      {
         termspace [ freetermspace ++ ] = f;
         remain += arities [f] - 1;
         rhs ++;  
      }
      else
      {
         for( i = 0; i < nrassignments; i ++ )
         {
            if( vars [i] == f )
               goto subst;
         }
         errorwithoutline( "unsound rewrite rule" ); 
      subst:
         (void) copyterm( vals [i] );
         remain -- ;
         rhs ++ ; 
      }
   }
   return 1;
}


#if 1 

static int copyruleterm( r )
int *r;
{
   int f, remain;
   int res;

   res = freetermspace;  
   remain = 1;
   while( remain > 0 ) 
   {
      f = * ( r ++ );
      termspace [ freetermspace ++ ] = f;
      if( f >= 0 )
         remain += arities [f] - 1;
      else
         remain -- ;
   }
   maketermends( res, freetermspace ); 
   return res;
}



static void printsetofrules ( s, lefts, rights )
char *s;
int lefts [] [ MAXRULESIZE ];
int rights [] [ MAXRULESIZE ];
{
   int i; 
   int t1, t2; 

   printstring( &outputfile, "Set of rules: " );
   printstring( &outputfile, s );
   printstring( &outputfile, "\n\n" );

   i = 0;
   while( lefts [i] [0] >= 0 )
   {
      printint( &outputfile, i );
      printstring( &outputfile, " :    "); 

      t1 = copyruleterm( lefts [i] ); 
      t2 = copyruleterm( rights [i] );
      printterm( &outputfile, t1 );
      printstring( &outputfile, " ==> " );
      printterm( &outputfile, t2 );
      printstring( &outputfile, "\n" );
      freetermspace = t1; 
      i ++ ;
   }
   printstring( &outputfile, "\n\n\n" );
}

#endif 


/*************************************************/
/* CONV_NNF: Rules for the negation normal form: */
/*************************************************/

static int nnfleft [] [ MAXRULESIZE ] = 
{
   { OP_IMPLIES, -1, -2 },
   { OP_EQUIV, -1, -2 },

   { OP_NOT, OP_AND, -1, -2 },
   { OP_NOT, OP_OR, -1, -2 },
   { OP_NOT, OP_FORALL, -1, -2 },
   { OP_NOT, OP_EXISTS, -1, -2 },
   { OP_NOT, OP_NOT, -1 },

   { OP_OR, -1, OP_FALSE },
   { OP_OR, OP_FALSE, -1 },
   { OP_AND, -1, OP_FALSE }, 
   { OP_AND, OP_FALSE, -1 }, 
   { OP_FORALL, -1, OP_FALSE }, 
   { OP_EXISTS, -1, OP_FALSE },  

   { OP_OR, -1, OP_TRUE },
   { OP_OR, OP_TRUE, -1 },
   { OP_AND, -1, OP_TRUE },
   { OP_AND, OP_TRUE, -1 },
   { OP_FORALL, -1, OP_TRUE },
   { OP_EXISTS, -1, OP_TRUE },

   { OP_AND, -1, OP_AND, -2, -3 },
   { OP_OR, -1, OP_OR, -2, -3 },

   { -1 } 
};
 
static int nnfright [] [ MAXRULESIZE ] = 
{
   { OP_OR, OP_NOT, -1, -2 },
   { OP_AND, OP_OR, OP_NOT, -1, -2, OP_OR, OP_NOT, -2, -1 }, 

   { OP_OR, OP_NOT, -1, OP_NOT, -2 },
   { OP_AND, OP_NOT, -1, OP_NOT, -2 },
   { OP_EXISTS, -1, OP_NOT, -2 },
   { OP_FORALL, -1, OP_NOT, -2 },
   { -1 },

   { -1 },
   { -1 },
   { OP_FALSE },
   { OP_FALSE },
   { OP_FALSE },
   { OP_FALSE }, 

   { OP_TRUE },
   { OP_TRUE },
   { -1 },
   { -1 },
   { OP_TRUE },
   { OP_TRUE },

   { OP_AND, OP_AND, -1, -2, -3 },
   { OP_OR, OP_OR, -1, -2, -3 } 
};


/***************************************************************/
/* CONV_SIMPQUANTFACT: Rules for pushing quantifiers into one  */
/* side of a binary operator, wherever possible.               */
/* Most rules are unsound and need additional checks.          */
/* The ones that are sound, are non-terminating.               */
/***************************************************************/

static int quantfactorleft [] [ MAXRULESIZE ] =
{
   { OP_FORALL, -1, OP_FORALL, -2, -3 },
   { OP_EXISTS, -1, OP_EXISTS, -2, -3 },

   { OP_FORALL, -1, OP_AND, -2, -3 },
   { OP_FORALL, -1, OP_AND, -2, -3 },
   { OP_FORALL, -1, OP_OR, -2, -3 },
   { OP_FORALL, -1, OP_OR, -2, -3 },
   { OP_FORALL, -1, OP_IMPLIES, -2, -3 },
   { OP_FORALL, -1, OP_IMPLIES, -2, -3 },

   { OP_EXISTS, -1, OP_AND, -2, -3 },
   { OP_EXISTS, -1, OP_AND, -2, -3 },
   { OP_EXISTS, -1, OP_OR, -2, -3 },
   { OP_EXISTS, -1, OP_OR, -2, -3 },
   { OP_EXISTS, -1, OP_IMPLIES, -2, -3 },
   { OP_EXISTS, -1, OP_IMPLIES, -2, -3 },

   { OP_FORALL, -1, -2 },
   { OP_EXISTS, -1, -2 },

   { -1 } 
};

 
static int quantfactorright [] [ MAXRULESIZE ] = 
{
   { OP_FORALL, -2, OP_FORALL, -1, -3 },
   { OP_EXISTS, -2, OP_EXISTS, -1, -3 },

   { OP_AND, OP_FORALL, -1, -2, -3 }, 
   { OP_AND, -2, OP_FORALL, -1, -3 },
   { OP_OR, OP_FORALL, -1, -2, -3 }, 
   { OP_OR, -2, OP_FORALL, -1, -3 },
   { OP_IMPLIES, OP_EXISTS, -1, -2, -3 }, 
   { OP_IMPLIES, -2, OP_FORALL, -1, -3 },

   { OP_AND, OP_EXISTS, -1, -2, -3 },
   { OP_AND, -2, OP_EXISTS, -1, -3 }, 
   { OP_OR, OP_EXISTS, -1, -2, -3 },
   { OP_OR, -2, OP_EXISTS, -1, -3 }, 
   { OP_IMPLIES, OP_FORALL, -1, -2, -3 },
   { OP_IMPLIES,  -2, OP_EXISTS, -1, -3 },

   { -2 },
   { -2 } 
};


/****************************************************************/
/* CONV_FACTOR:   Rules for factoring into clausal normal form  */
/****************************************************************/


static int factorleft [] [ MAXRULESIZE ] =
{
   { OP_OR, OP_AND, -1, -2, -3 },
   { OP_OR, -1, OP_AND, -2, -3 },
   { OP_OR, -1, OP_OR, -2, -3 },
   { OP_FORALL, -1, OP_AND, -2, -3 },
   { -1 } 
};

static int factorright [] [ MAXRULESIZE ] =
{
   { OP_AND, OP_OR, -1, -3, OP_OR, -2, -3 }, 
   { OP_AND, OP_OR, -1, -2, OP_OR, -1, -3 }, 
   { OP_OR, OP_OR, -1, -2, -3 }, 
   { OP_AND, OP_FORALL, -1, -2, OP_FORALL, -1, -3 } 
};



/* Make one sweep of applications of the rewrite rules lhs[i], rhs[i]. */
/* The result is stored on termspace [ freetermspace ], also if no     */
/* replacement was made. In the case that a replacement was made 1 is  */
/* returned.                                                           */

static int onerewritepass( startofinput, endofinput, lhs, rhs, 
                           conversion )
int startofinput, endofinput;
int lhs [] [ MAXRULESIZE ], rhs [] [ MAXRULESIZE ];
int conversion; 
{
   int rewr;
   int i;
   int r;
   int resultbase, res;

   rewr = 0;
   resultbase = freetermspace; 

   i = startofinput;
   while( i < endofinput )
   {
      for( r = 0; lhs [r][0] >= 0; r ++ )
      {
         res = freetermspace;
         if( applyrewriterule( i, lhs [r], rhs [r], 0, NULL ))
         {
            storesimprepl( conversion, r, resultbase, res );
            i = termends [ i ];

            rewr = 1;
            goto next;
         }
      }
      termspace [ freetermspace ++ ] = termspace [ i ++ ];
   next:
      if( freetermspace >= MAXTERMSPACE - 10 ) 
         errorwithoutline( "termspace is full" ); 
   }
   return rewr;
}


/**********************************************************************/
/* convquantfactor tries to push quantifiers into one side of binary  */
/* operators.                                                         */
/**********************************************************************/

/* Rewrite conditions. */

static int qf_exchange( v1, v2, v3, v4 )
int v1, v2, v3, v4;
{
   int left, right;
   int a1, a2;

   while( termspace [ v3 ] == OP_FORALL || termspace [ v3 ] == OP_EXISTS ) 
      v3 += 2;
   if( termspace [ v3 ] != OP_AND && termspace [ v3 ] != OP_OR && 
       termspace [ v3 ] != OP_IMPLIES ) 
      return 0;

   left = v3 + 1;
   right = termends [ left ];
   a1 = termspace [ v1 ];
   a2 = termspace [ v2 ];

   if( !fovarfree( a2, left ) || !fovarfree( a2, right ))
      return 0;
   if( fovarfree( a1, left )  && fovarfree( a1, right ))
      return 0;
   return 1;   
}


static int qf_totheleft( v1, v2, v3, v4 )
int v1, v2, v3, v4;
{
   int a;

   a = termspace [ v1 ];
   if( !fovarfree( a, v3 ))
      return 1;
   return 0;   
}


static int qf_totheright( v1, v2, v3, v4 )
int v1, v2, v3, v4;
{
   int a;

   a = termspace [ v1 ];
   if( !fovarfree( a, v2 ))
      return 1;
   return 0;
}


static int qf_notfree( v1, v2, v3, v4 )
int v1, v2, v3, v4;
{
   int a;

   a = termspace [ v1 ];
   if( !fovarfree( a, v2 ))
      return 1;
   return 0;
}

 
static int convquantfactor( startofinput, endofinput, conversion )
int startofinput, endofinput;
int conversion;
{
   int rewr;
   int i;
   int r;
   int resultbase, res;
   int (*condition)(int, int, int, int );

   if( conversion != CONV_SIMPQUANTFACT ) exit(0);
 
   rewr = 0;
   resultbase = freetermspace;
 
   i = startofinput;
   while( i < endofinput )
   {
      for( r = 0; quantfactorleft [r][0] >= 0; r ++ )
      {
         res = freetermspace;
         /* Select the condition. */
         switch(r)
         {
         case 0:
         case 1:
            condition = qf_exchange; 
            break;
         case 2:
         case 4:
         case 6:
         case 8:
         case 10:
         case 12:
            condition = qf_totheleft;
            break;
         case 3:
         case 5:
         case 7:
         case 9:
         case 11:
         case 13:
            condition = qf_totheright; 
            break;
         case 14:
         case 15:
            condition = qf_notfree;
            break;
         default:
            exit(0);
         } 
         if( applyrewriterule( i, quantfactorleft [r], 
                               quantfactorright [r], 1, condition ))
         {
            storesimprepl( conversion, r, resultbase, res );
            i = termends [i];
 
            rewr = 1;
            goto next;
         }
      }
      termspace [ freetermspace ++ ] = termspace [ i ++ ];
   next:
      if( freetermspace >= MAXTERMSPACE - 10 )
         errorwithoutline( "termspace is full" );
   }
   return rewr;
}



/*****************************************************************/
/*   Produce an initial clause from a literal list.              */
/*   makeinitialclause assumes very strongly that the            */
/*   literal list has been typechecked already.                  */
/*****************************************************************/

/* Translate a literal from a listclause.   */

static void translistliteral( literal )
int literal;
{
   int f;
   int ind;
   int i;

   if( freetermspace >= MAXTERMSPACE - 10 ) 
      errorwithoutline( "termspace is full" );

   if( termspace [ literal ] == OP_NOT ) 
   {
      termspace [ freetermspace ++ ] = OP_NOT;
      literal ++ ;
   }
   if( termspace [ literal ] == OP_NOT ) exit(0);

   termspace [ freetermspace ++ ] = termspace [ literal ];
   for( ind = literal + 1; ind < termends [ literal ]; ind ++ )
   {
      f = termspace [ ind ];
      if( isprologvar( f ) && arities [f] == 0 )
      {
         for( i = 0; i < freerenaming; i ++ )
         {
            if( renaming [i]. var == f )
               goto found;
         }
         renaming [ i ]. var = f;
         renaming [ i ]. varorig = 0; 
         if( i != freerenaming ) exit(0);
         freerenaming ++ ;
         if( freerenaming >= MAXNRVARS - 3 ) 
            errorwithoutline( "too many variables in clause" );            
      found:
         termspace [ freetermspace ++ ] = - ( 1 + i );
      }
      else
      {
         termspace [ freetermspace ++ ] = f;
      }
      if( freetermspace >= MAXTERMSPACE - 10 ) 
         errorwithoutline( "termspace is full" );
   }       
}


/* translate a clause that it represented by a list. */ 

static void translistclause( term )
int term;
{
   freerenaming = 0;
   while( termspace [ term ] == OP_DOT ) 
   {
      term ++ ; 
      termspace [ freetermspace ++ ] = OP_DOT;
      translistliteral( term );
      term = termends [ term ];
   }
   if( termspace [ term ] != OP_NIL ) exit(0);
   termspace [ freetermspace ++ ] = OP_NIL;
   return;
}



static void transfoliteral( literal )
int literal;
{
   int f;
   int ind;
   int i;

   if( freetermspace >= MAXTERMSPACE - 10 ) 
      errorwithoutline( "termspace is full" );

   if( termspace [ literal ] == OP_NOT )
   {
      termspace [ freetermspace ++ ] = OP_NOT;
      literal ++ ;
   }
   if( termspace [ literal ] == OP_NOT ) exit(0);
 
   termspace [ freetermspace ++ ] = termspace [ literal ];
   for( ind = literal + 1; ind < termends [ literal ]; ind ++ )
   {
      f = termspace [ ind ];
      if( arities [f] == 0 )
      {
         for( i = freerenaming - 1; i >= 0; i -- )
         {
            if( renaming [i]. var == f &&  
                renaming [i]. varorig <= ind && 
                termends [ renaming [i]. varorig ] > ind )
            {
               termspace [ freetermspace ++ ] = - ( 1 + i );
               goto next;
            }
         }
         if( isprologvar( f ))
            exit(0);   /* should not have passed the type check. */
         else
            termspace [ freetermspace ++ ] = f;
      }
      else
         termspace [ freetermspace ++ ] = f;
   next:
      if( freetermspace >= MAXTERMSPACE - 10 ) 
         errorwithoutline( "termspace is full" );
   }     
}
 

/* translate a clause that is represented by a first order formula. */  

static void transfoclause( term )
int term;
{
   int i;

   /* collect the variables.  */
 
   freerenaming = 0;
   for( i = term; i < termends [ term ]; i ++ )
   {
      if( termspace [i] == OP_EXISTS || termspace [i] == OP_AND || 
          termspace [i] == OP_IMPLIES || termspace [i] == OP_EQUIV )
         exit(0);

      if( termspace [i] == OP_FORALL ) 
      {
         renaming [ freerenaming ]. var = termspace [ i + 1 ];
         renaming [ freerenaming ]. varorig = i;
         if( freerenaming >= MAXNRVARS - 3 ) 
            errorwithoutline( "too many variables in clause" );
         freerenaming ++ ;
      }   
   }

   i = term;
   while( i < termends [ term ] )
   {
      if( termspace [i] == OP_FORALL )
         i += 2;
      else
      { 
         if( termspace [i] == OP_OR ) 
            i ++ ; 
         else
         {
            termspace [ freetermspace ++ ] = OP_DOT;

            transfoliteral( i );
            i = termends [i];
         }
      }
   } 
   termspace [ freetermspace ++ ] = OP_NIL;
} 


/* make a clause from term.       */   

static void makeclause ( term )
int term;
{
   if( termspace [ term ] == OP_NIL || termspace [ term ] == OP_DOT )
      translistclause( term ); 
   else
      transfoclause( term );
}


static void maketheclauses( t1, t2 )
int t1, t2;
{
   int i;

   for( i = t1; i < t2; i = termends [i] )
      makeclause( i );
}
 
 
/***********************************************************************/
/* Convert the formulae in termspace [ t1 .. t2 ] into clauses, using  */
/* the transformation specified by fullconversion.                     */
/* If the formulae in termspace [ t1 .. t2 ] are already clauses,      */
/* then no conversion is made.                                         */
/* A record in translations [] is created, where the result is         */
/* stored. axioms indicates whether or not termspace [ t1 .. t2 ] is   */
/* the set of axioms, or a goal. Nothing is done with this             */
/* information, but it is stored in translations []. axioms.           */
/* If totalproof == 1, a detailed record of the replacement is kept.   */
/* makeconversion returns the index in translations where it stored    */
/* the result.                                                         */
/***********************************************************************/


int makeconversion( t1, t2, fullconversion )
int t1, t2;
int fullconversion; 
{
   int *translation;
   int lastcopy;
   int i, j;
   int rewr;
   int last;
   char *s;
   int nrpasses;

   /* If all formulae are clauses, then we don't translate.   */

   rewr = 0;
   for( i = t1; i < t2; i = termends [i] )
   {
      if( ! hastype( i, TYPE_CLAUSE ))
         rewr = 1;
   }

   if( rewr == 0 )
   {
      fullconversion = FULLCONV_NONE;
   }
   else
   {
      /* If one of the formulae is a clause, or modal, or ill-typed, */
      /* then quit with an appropriate message.                      */

      for( i = t1; i < t2; i = termends [i] )
      {
         if( ! hastype( i, TYPE_FOFORM ))
         {
            if( hastype( i, TYPE_MODFORM )) 
               printstring( &outputfile,
                     "Bliksem does not support modal formulae anymore: " );
            else
               printstring( &outputfile, "Input is ill-typed: " );
            printterm( &outputfile, i );
            errorwithoutline( " " );
         }
      }
   } 

   for( i = 0; i < nrtranslations; i ++ )
   {
      if( translations [i]. from1 == t1 &&
          translations [i]. from2 == t2 &&
          translations [i]. fullconv == fullconversion )
      {
         return i;
      }
   }

   printstring( &outputfile, "Conversion number ");
   printint( &outputfile, nrtranslations + 1 );
   printstring( &outputfile, ", " );
   printfullconversion( &outputfile, fullconversion );
   printstring( &outputfile, ".\n\n" );

   if( nrtranslations >= MAXTRANSLATIONS - 2 ) 
      errorwithoutline( "too many translations" );

   translations [ nrtranslations ]. fullconv = fullconversion;
   translations [ nrtranslations ]. from1 = t1;
   translations [ nrtranslations ]. from2 = t2;
   translations [ nrtranslations ]. startofresult = freetermspace;
   translations [ nrtranslations ]. startofjust = freejustifications;  

   translation = expandfullconv( fullconversion );
 
   /* make a fresh copy: */

   if( freetermspace + ( t2 - t1 ) + 10 >= MAXTERMSPACE ) 
      errorwithoutline( "termspace is full" );

   lastcopy = freetermspace;
   for( i = t1; i < t2; i ++ )
      termspace [ freetermspace ++ ] = termspace [i];
   maketermends( lastcopy, freetermspace );

   while( *translation >= 0 )
   {
      nrpasses = 0;
      rewr = 1;
      while( rewr )
      {
         last = freetermspace; 
         switch( *translation )
         {
            case CONV_NNF:
               rewr = onerewritepass( lastcopy, freetermspace,
                                      nnfleft, nnfright, CONV_NNF ); 
               break;
            case CONV_SKOLEMFREE:
            case CONV_SKOLEM:
            case CONV_SKOLEMNG:
               rewr = convskolem( lastcopy, freetermspace, *translation ); 
               break;
            case CONV_STRUCTNA1:
            case CONV_STRUCTNA2:
            case CONV_STRUCTFACT1:
            case CONV_STRUCTFACT2:
            case CONV_STRUCTFACTOS1:
            case CONV_STRUCTFACTOS2:
            case CONV_STRUCTGUARD:
               rewr = convstruct( lastcopy, freetermspace, *translation );
               break;
            case CONV_SIMPQUANTFACT:
               rewr = convquantfactor( lastcopy, freetermspace, 
                                       CONV_SIMPQUANTFACT ); 
               break;
            case CONV_SIMPANTIPRENEX:
               rewr = convantiprenex( lastcopy, freetermspace, 
                                      CONV_SIMPANTIPRENEX ); 
               break;
            case CONV_FACTOR:
               rewr = onerewritepass( lastcopy, freetermspace,
                                      factorleft, factorright, CONV_FACTOR );
               if( ! rewr )
               {
                  freetermspace = last;
                  rewr = convfactor( lastcopy, freetermspace, *translation );
               }
               break;
            default:
               exit(0);
         }
         if( rewr )
         {
            j = lastcopy;
            for( i = last; i < freetermspace; i ++ )
               termspace [ j ++ ] = termspace [ i ];
            freetermspace = j;
            maketermends( lastcopy, freetermspace ); 
            nrpasses ++ ;
         }
         else
            freetermspace = last; 
      }

      switch( *translation )
      {
      case CONV_NNF:
         s = "Nnf:"; 
         break;
      case CONV_SKOLEMFREE:
      case CONV_SKOLEM:
      case CONV_SKOLEMNG:
         s = "Skolemization:"; 
         break;
      case CONV_STRUCTNA1:
      case CONV_STRUCTFACT1:
      case CONV_STRUCTFACTOS1:
      case CONV_STRUCTGUARD:
         s = "Structural:"; 
         break;
      case CONV_STRUCTNA2:
      case CONV_STRUCTFACT2:
      case CONV_STRUCTFACTOS2:
         if( nrpasses > 0 )
            s = "Structural:";
         else
            s = (char*) NULL;
         break;
      case CONV_SIMPQUANTFACT:
         if( nrpasses > 0 )
            s = "Push quantifiers inward:";
         else
            s = (char*) NULL;
         break;
      case CONV_SIMPANTIPRENEX:
         if( nrpasses > 0 )
            s = "Antiprenex:"; 
         else
            s = (char*) NULL;
         break;
      case CONV_FACTOR:
         s = "Factor:"; 
         break;
      default:
         exit(1);
      }
/*----Turned off for CASC
 *    if( s != (char*) NULL ) 
 *       printtermlist( &outputfile, s, lastcopy, freetermspace ); 
 */
      translation ++ ;
   }

   last = freetermspace;
   maketheclauses( lastcopy, freetermspace );

   j = lastcopy;
   for( i = last; i < freetermspace; i ++ )
      termspace [ j ++ ] = termspace [i];
   freetermspace = j;
   maketermends( lastcopy, freetermspace );

/*----Turned off for CASC
 * printtermlist( &outputfile, "Clauses:", lastcopy, freetermspace );
 */
 
   translations [ nrtranslations ]. endofresult = freetermspace;
   translations [ nrtranslations ]. endofjust = freejustifications;
   return ( nrtranslations ++ ) ; 
}


