

/******************************************************************/
/* file orders.c contains term and atom comparison functions.     */
/* H. de Nivelle, 1998.                                           */
/******************************************************************/

#define CLASS extern

#include "basics.h" 
#include "extfunc.h" 
#include "resolution.h"


/* The following datastructure is used for the KBO, for NLO's, and  */
/* for the <_d order.                                               */

static struct
{
   int var1;
   int var2; 
   int depthfreq;
}
   varprofile [ MAXNRVARS ];
static int freevarprofile;


#if 0

static void printvarprofile( )
{
   int i;

   printstring( &outputfile, "Varprofile:\n\n");
   for( i = 0; i < freevarprofile; i ++ ) 
   {
      printvariable( &outputfile, varprofile [i]. var1 );
      printstring( &outputfile, " " ); 
#if 0
      printvariable( &outputfile, varprofile [i]. var2 ); 
      printstring( &outputfile, " " ); 
#endif
      printint( &outputfile, varprofile [i]. depthfreq );
      printstring( &outputfile, "\n"); 
   }
   printstring( &outputfile, "\n\n");
}

#endif


/****************************************************************/
/*                Print a comparison.                           */
/****************************************************************/

static void printcomparison( f, c )
bliksemfile *f;
int c;
{
   switch( c )
   {
   case COMP_LESS:
      printstring( f, "less than" ); 
      return; 
   case COMP_EQUIV:
      printstring( f, "equivalent with" );
      return; 
   case COMP_MORE:
      printstring( f, "more than" );
      return; 
   case COMP_UNCOMPARABLE:
      printstring( f, "uncomparable with" );
      return;
   default:
      internalproblem( "dont know comparison" );
   }
   return; 
}



/*****************************************************************/
/*      Complexity measures on terms and on clauses:             */
/*****************************************************************/

/* compute the depth of term x.   */ 

static int termdepth( t )
int t;
{
   int op, arity;
   int i;
   int d, depth;

   op = termspace [t];
   if( op < 0 ) 
      return 1;
   else
   { 
      depth = 0;
      arity = arities [ op ];
      t = t + 1; 
 
      for( i = 0; i < arity; i ++ ) 
      {
         d = termdepth( t );
         if( d > depth )
            depth = d;
         t = termends [t];  
      }
      return depth + 1;
   }
}
 

/* Compute the depth of a literal. The negation symbol is not counted, */
/* but the predicate symbol is. litdepth is called only by             */
/* clausedepth.                                                        */

static int litdepth( lit )
int lit;
{
   if ( termspace [ lit ] == OP_NOT ) 
      lit = lit + 1; 
   return termdepth( lit );
}
  
 
/* compute the depth of a clause. The depth of the empty clause is 0. */

int clausedepth( clause )
int clause;
{
   int depth, d;
   int i;

   depth = 0;
   for( i = clauses [ clause + CLAUSE_LITERALS ];
        i < clauses [ clause + CLAUSE_END ];
        i ++ )
   {
      d = litdepth( clauses [i] );
      if( d > depth )
         depth = d;
   }
   return depth;
}


/****************************************************************/
/* Vardepth is the deepest depth at which a variable occurs,    */
/* or -1 if there is no variable.                               */
/****************************************************************/


int termvardepth( x )
int x;
{
   int op, arity;
   int i;
   int d, depth;

   op = termspace [x];
   if( op < 0 ) 
   {
      return 0;
   }
   else
   {
      depth = -1;
      arity = arities [ op ]; 
      x = x + 1;

      for( i = 0; i < arity; i ++ ) 
      {
         d = termvardepth( x );
         if( d > depth )
            depth = d;
         x = termends [ x ];
      }
      if( depth < 0 ) 
         return -1;
      else
         return 1 + depth;
   }
}


/****************************************************************/
/* Weight is used for selecting clauses. Complexity is used     */
/* for sorting clauses. These are different things.             */
/****************************************************************/


/* Weight of a literal.  */

static int litweight( lit )
int lit;
{
   int weight;
   int end;
   int op;

   weight = 0;
   end = termends [ lit ];
   while( lit < end )
   {
      op = termspace [ lit ++ ];
      if( op >= 0 )
         weight += operatorweights [ op ];
      else
         weight += 1;
   }
   return weight;
}


/* weight of a clause.  */

int clauseweight( clause )
int clause;
{
   int weight;
   int i;
  
   weight = 0; 
   for( i = clauses [ clause + CLAUSE_LITERALS ];
        i < clauses [ clause + CLAUSE_END ];
        i ++ )
   {
      weight += litweight( clauses [i] );
   }
   return weight;
}


/******************************************************************/
/* Compute complexity of term, using operatororder [].            */
/******************************************************************/

static int termcompl( term )
int term;
{
   int i;
   int f;
   int compl;

   compl = 0;
   for( i = term; i < termends [ term ]; i ++ )
   {
      f = termspace [i];
      if( f < 0 )
         compl ++ ;
      else
         compl += operatororder [f];
   } 
   return compl;
}


/******************************************************/
/* count the number of distinct variables in term x.  */
/******************************************************/


int termnrvars( x )
int x;
{
   int i, j;
   int xend;
   int v;

   xend = termends [ x ]; 

   freevarprofile = 0;
   for( i = x; i < xend; i ++ ) 
   {
      v = termspace [i]; 
      if( v < 0 ) 
      {
         for( j = 0; j < freevarprofile; j ++ ) 
         {
            if( varprofile [j]. var1 == v )
               goto found;
         }
         varprofile [ freevarprofile ++ ]. var1 = v;
         /* no test is necessary since the number of variables is */
         /* limited to MAXNRVAR.                                  */
      found:
         ; 
      }
   }
   return freevarprofile; 
}


/* Produce number of variables in clause.   */
 
int clausenrvars( clause )
int clause;
{
   int cl; 
   int t, tend;
   int a;
   int i; 

   freevarprofile = 0;

   for( cl = clauses [ clause + CLAUSE_LITERALS ];
        cl < clauses [ clause + CLAUSE_END ];
        cl ++ ) 
   {
      t = clauses [cl];
      tend = termends [t];

      while( t < tend )
      {
         a = termspace [t]; 
         if( a < 0 )
         {
            for( i = 0; i < freevarprofile; i ++ )
            {
               if( varprofile [i]. var1 == a )
                  goto found;
            }
            varprofile [ freevarprofile ++ ]. var1 = a;
         }
      found: 
         t ++ ; 
      }
   }
   return freevarprofile; 
}


/***********************************************************************/
/* Compute the lexicographic extension of order on the argument lists  */
/* of terms t1 and t2 comparing from left to right.                    */
/***********************************************************************/


static int lr_lexext( t1, t2, order )
int t1;
int t2;
int order( int, int ); 
{
   int sub1, sub2;
   int i;
   int arity1, arity2;
   int cmp;

#if 0
   printstring( &outputfile, "LR-lexicographic ext: " );
   printterm( &outputfile, t1 );
   printstring( &outputfile, " == " );
   printterm( &outputfile, t2 );
   printstring( &outputfile, "\n" );
#endif

   if( termspace [ t1 ] < 0 || termspace [ t2 ] < 0 )
      internalproblem( "argument to lr_lexext should not be variable" );

   arity1 = arities [ termspace [ t1 ]];
   arity2 = arities [ termspace [ t2 ]];

   sub1 = t1 + 1; 
   sub2 = t2 + 1;
   for( i = 0; i < arity1 && i < arity2; i ++ )
   {
      cmp = order( sub1, sub2 );
      if( cmp != COMP_EQUIV ) 
         return cmp;
      sub1 = termends [ sub1 ];
      sub2 = termends [ sub2 ];
   }
   if( arity1 < arity2 ) 
      return COMP_LESS;
   if( arity1 > arity2 )
      return COMP_MORE;
   return COMP_EQUIV;
}



/*****************************************************************/
/* Compute the lexicographic extension of order on t1 and t2,    */
/* comparing from right to left.                                 */
/*****************************************************************/


static int rl_lexext( t1, t2, order )
int t1;
int t2;
int order( int, int );
{
   int sub1, sub2;
   int i, j;
   int arity1, arity2;
   int cmp;

#if 0
   printstring( &outputfile, "RL-lexicographic ext: " );
   printterm( &outputfile, t1 );
   printstring( &outputfile, " == " );
   printterm( &outputfile, t2 );
   printstring( &outputfile, "\n" );
#endif
   if( termspace [ t1 ] < 0 || termspace [ t2 ] < 0 )
      internalproblem( "argument to rl_lexext should not be variable" );

   arity1 = arities [ termspace [ t1 ]];
   arity2 = arities [ termspace [ t2 ]];
   
   for( i = 0; i < arity1 && i < arity2; i ++ )
   {
      sub1 = t1 + 1;
      for( j = 1; j < arity1 - i; j ++ )
         sub1 = termends [ sub1 ];

      sub2 = t2 + 1;
      for( j = 1; j < arity2 - i; j ++ )
         sub2 = termends [ sub2 ];

      cmp = order( sub1, sub2 );
      if( cmp != COMP_EQUIV ) 
         return cmp;
   }

   if( arity1 < arity2 )
      return COMP_LESS;
   if( arity1 > arity2 )
      return COMP_MORE;
   return COMP_EQUIV;
}


/**************************************************************************/
/* Compute the multiset extension of order on the arguments of t1 and t2. */
/**************************************************************************/


static int multisetext( t1, t2, order )
int t1;
int t2;
int order( int, int ); 
{
   int sub1, sub2;
   int i, j;
   int arity1, arity2;
   int inset1 [ MAXARITY ], inset2 [ MAXARITY ];
   int cmp;

#if 0
   printstring( &outputfile, "multisetext: " );
   printterm( &outputfile, t1 );
   printstring( &outputfile, " == " );
   printterm( &outputfile, t2 );
   printstring( &outputfile, "\n" );
#endif
   if( termspace [ t1 ] < 0 || termspace [ t2 ] < 0 )
      internalproblem( "argument to multisetext should not be variable" );

   arity1 = arities [ termspace [ t1 ]];
   arity2 = arities [ termspace [ t2 ]];

   if( arity1 >= MAXARITY || arity2 >= MAXARITY ) 
   {
      errorwithoutline( "arity is too high in multiset comparison" );
   }

   for( i = 0; i < arity1; i ++ )
   {
      inset1 [i] = 1;
   }
   for( j = 0; j < arity2; j ++ )
   {
      inset2 [j] = 1;
   }

   sub1 = t1 + 1; 
   for( i = 0; i < arity1; i ++ )
   {
      sub2 = t2 + 1;
      for( j = 0; j < arity2; j ++ )
      {
         if( inset2 [j] )
         {
            if( order( sub1, sub2 ) == COMP_EQUIV ) 
            {
               inset1[i] = 0;
               inset2[j] = 0;
               goto foundequiv;
            }
         }
         sub2 = termends [ sub2 ];
      }
   foundequiv:
      sub1 = termends [ sub1 ];
   }
   for( i = 0; i < arity1; i ++ )
   {
      if( inset1 [i] ) 
         goto notequivalent;
   }
   for( j = 0; j < arity2; j ++ )
   {
      if( inset2 [j] )
         goto notequivalent;
   }
   return COMP_EQUIV;
  
notequivalent:
   sub1 = t1 + 1;  
   for( i = 0; i < arity1; i ++ )
   {
      if( inset1 [i] )
      {
         sub2 = t2 + 1;
         for( j = 0; j < arity2; j ++ )
         {
            if( inset2 [j] )
            {
               cmp = order( sub1, sub2 );
               if( cmp == COMP_LESS ) 
               {
                  inset1 [i] = 0;
                  goto deleted1;
               }
               if( cmp == COMP_MORE ) 
               {
                  inset2 [j] = 0;
                  goto deleted2;
               }
            }
         deleted2:
            sub2 = termends [ sub2 ];
         }
      }
   deleted1:
      sub1 = termends [ sub1 ];
   }

   sub1 = 0;
   for( i = 0; i < arity1; i ++ )
   {
      if( inset1 [i] )
         sub1 ++ ;
   }
   sub2 = 0;
   for( j = 0; j < arity2; j ++ ) 
   {
      if( inset2 [j] )
         sub2 ++ ;
   }
   if( sub1 == 0 && sub2 != 0 )
      return COMP_LESS; 
   if( sub1 != 0 && sub2 == 0 )
      return COMP_MORE;
   return COMP_UNCOMPARABLE;
}



/****************************************************************/
/*                The <_d order from FLTZ93.                    */
/****************************************************************/


/* Produce a list of variables in term, and how deep they occur.  */

static void depthprofile( term, depth )
int term;
int depth;
{
   int op, i;

   op = termspace [ term ];
   if( op < 0 )
   {
      for( i = 0; i < freevarprofile; i ++ )
      {
         if( varprofile [i]. var1 == op &&
             varprofile [i]. depthfreq < depth )
         {
            varprofile [i]. depthfreq = depth;
            return;
         }
      }
      varprofile [ freevarprofile ]. var1 = op;
      varprofile [ freevarprofile ]. depthfreq = depth;
      freevarprofile ++ ;
      return;
   }
   term ++ ;
   for( i = 0; i < arities [ op ]; i ++ )
   {
      depthprofile( term, depth + 1 );
      term = termends [ term ];
   }
   return;

}


static int withindepthprofile( term, depth )
int term;
int depth;
{
   int op, i;

   op = termspace [ term ];
   if( op < 0 )
   {
      for( i = 0; i < freevarprofile; i ++ )
      {
         if( varprofile [i]. var1 == op )
         { 
            if( depth >= varprofile [i]. depthfreq )
               return 0;
            return 1;
         }
      }
      return 0;
   }
   term ++ ;
   for( i = 0; i < arities [ op ]; i ++ )
   {
      if( !withindepthprofile( term, depth + 1 ))
         return 0;
      term = termends [ term ];
   }
   return 1;
}


/* The <_d ordering from FLTZ93.         */          

static int ord_d( t1, t2 )
int t1, t2;
{
   int d1, d2;

   d1 = termdepth( t1 );
   d2 = termdepth( t2 );

   if( d1 < d2 )
   {
      freevarprofile = 0;
      depthprofile( t2, 0 );
      if( withindepthprofile( t1, 0 ))
         return COMP_LESS;
      return COMP_UNCOMPARABLE;
   }

   if( d1 > d2 )
   {
      freevarprofile = 0;
      depthprofile( t1, 0 );
      if( withindepthprofile( t2, 0 )) 
         return COMP_MORE;
      return COMP_UNCOMPARABLE;
   }

   freevarprofile = 0;
   depthprofile( t1, 0 );
   if( withindepthprofile( t2, 0 ))
   {
      freevarprofile = 0;
      depthprofile( t2, 0 );
      if( withindepthprofile( t1, 0 ))
         return COMP_EQUIV;
   }
   return COMP_UNCOMPARABLE;   
}


/***************************************************************/
/*                 The Knuth-Bendix order.                     */
/***************************************************************/


/* Every variable occurring in t1 occurs at least as often in t2. */

static int varmultisubset( t1, t2 )
int t1;
int t2;
{
   int v;
   int i, j;
   int t1end, t2end;   

   freevarprofile = 0;

   t2end = termends [ t2 ]; 
   for( i = t2; i < t2end; i ++ ) 
   {
      v = termspace [i];
      if( v < 0 ) 
      {
         for( j = 0; j < freevarprofile; j ++ ) 
         {
            if( varprofile [j]. var1 == v ) 
            {
               varprofile [j]. depthfreq ++ ; 
               goto foundt2;
            }
         }
         varprofile [ freevarprofile ]. var1 = v;
         varprofile [ freevarprofile ]. depthfreq = 1;
         freevarprofile ++ ; 
      foundt2:
         ;
      }
   }

   t1end = termends [ t1 ]; 
   for( i = t1; i < t1end; i ++ ) 
   {
      v = termspace [i];
      if( v < 0 ) 
      {
         for( j = 0; j < freevarprofile; j ++ ) 
         {
            if( varprofile [j]. var1 == v ) 
            {
               varprofile [j]. depthfreq -- ; 
               if( varprofile [j]. depthfreq < 0 ) 
                  return 0;
               goto foundt1; 
            }
         }
         return 0;
      foundt1:
         ;
      }
   }

   return 1;
}


static int kbo( term1, term2 )
int term1;
int term2;
{
   int d1, d2;
   int f1, f2;
   int dir;

#if 0
   printstring( &outputfile, "kbo: " );
   printterm( &outputfile, term1 );
   printstring( &outputfile, " ; " );
   printterm( &outputfile, term2 );
   printstring( &outputfile, "\n" );
#endif 

#if 0
   d1 = termcompl( term1 );  /* I'm not sure what complexity should */
   d2 = termcompl( term2 );  /* be used here.                       */
#else
   d1 = termends [ term1 ] - term1;
   d2 = termends [ term2 ] - term2;
#endif

   if( d1 < d2 )
   {
      if( varmultisubset( term1, term2 ))
         return COMP_LESS;
      return COMP_UNCOMPARABLE;
   }
   if( d1 > d2 )
   {
      if( varmultisubset( term2, term1 ))
         return COMP_MORE;
      return COMP_UNCOMPARABLE;
   }

   f1 = termspace [ term1 ];
   f2 = termspace [ term2 ];

   if( f1 < 0 )
   {
      if( f2 < 0 )
      {
         if( f1 == f2 )
            return COMP_EQUIV;
         else
            return COMP_UNCOMPARABLE;
      }
      else
         return COMP_UNCOMPARABLE;
   }
   else
   {
      if( f2 < 0 )
         return COMP_UNCOMPARABLE;
   }

   if( varmultisubset( term1, term2 ) && varmultisubset( term2, term1 ))
   {
      d1 = operatororder [ f1 ];
      d2 = operatororder [ f2 ];

      if( d1 == d2 )
      {
         dir = operatorassoc [ f1 ];
         if( dir != operatorassoc [ f2 ] )
            return COMP_EQUIV;

         if( dir < 0 )
            return lr_lexext( term1, term2, kbo ); 
         if( dir == 0 )
            return multisetext( term1, term2, kbo );
         if( dir > 0 )
            return rl_lexext( term1, term2, kbo );
      }
      if( d1 < d2 )
         return COMP_LESS;
      else
         return COMP_MORE;
   }
   return COMP_UNCOMPARABLE;
}



/**********************************************************************/
/* Combination of lexicographic path order and recursive path order.  */
/* It can be specified by operatorassoc [] how the arguments of       */
/* each operator should be treated. This is often called              */
/* 'recursive path order with status'.                                */
/**********************************************************************/



/* Return 1 if all arguments of t1 are smaller than t2 under     */
/* order.                                                        */

static int allargumentsless( t1, t2, order )
int t1; 
int t2;
int order( int, int );
{
   int sub1, endofterm1;
   int cmp;

#if 0
   printstring( &outputfile, "allargumentsless: " );
   printterm( &outputfile, t1 );
   printstring( &outputfile, ", " );
   printterm( &outputfile, t2 );
   printstring( &outputfile, "\n" );
#endif

   endofterm1 = termends [ t1 ];
   for( sub1 = t1 + 1; sub1 < endofterm1; sub1 = termends [ sub1 ] )
   {
      cmp = order( sub1, t2 );
      if( cmp != COMP_LESS ) 
         return 0;
   }
   return 1;
}


/* Return 1 if one argument of t1 is greater than or equal to */
/* t2 under order.                                            */


static int oneargumentgeq( t1, t2, order )
int t1; 
int t2;
int order( int, int );
{
   int sub1, endofterm1;
   int cmp;

#if 0
   printstring( &outputfile, "oneargumentgeq: " );
   printterm( &outputfile, t1 );
   printstring( &outputfile, ", " );
   printterm( &outputfile, t2 );
   printstring( &outputfile, "\n" );
#endif

   endofterm1 = termends [ t1 ];
   for( sub1 = t1 + 1; sub1 < endofterm1; sub1 = termends [ sub1 ] )
   {
      cmp = order( sub1, t2 );
      if( cmp == COMP_MORE || cmp == COMP_EQUIV ) 
         return 1;
   }
   return 0;
}


/* Since the lrpo compares its subterms too often, we need       */
/* to do caching.                                                */

#define MAXLRPOCACHE 10000
static struct 
{
   int t1;
   int t2;
   int cmp;
}
   lrpocache [ MAXLRPOCACHE ];
static int freelrpocache;


/* Return 1 if symbol f occurs in term.               */

static int symboloccurs( f, term )
int f;
int term;
{
   int endofterm;

   endofterm = termends [ term ];
   while( term < endofterm )
   {
      if( termspace [ term ] == f )
         return 1;
      term ++ ;
   }
   return 0;
}



/* Compute the lrpo on terms t1 and t2. Decide for each symbol,  */
/* whether or not it is a RPO, LPO(LR) or LPO(RL).               */
/* This is also sometimes called, RPO with status.               */

static int lrpo( t1, t2 )
int t1, t2;
{
   int sub1, sub2;
   int end1, end2;
   int cmp;

   int f1, f2;
   int dir1, dir2; 
   int i;

#if 0 
   printstring( &outputfile, "LRPO( " );
   printterm( &outputfile, t1 );
   printstring( &outputfile, " == " );
   printterm( &outputfile, t2 );
   printstring( &outputfile, " )\n" );
#endif

   /* Try to retrieve the value from the cache.   */

   for( i = 0; i < freelrpocache; i ++ )
   {
      if( lrpocache [i]. t1 == t1 && lrpocache [i]. t2 == t2 )
      {
         return lrpocache [i]. cmp;
      }

      if( lrpocache [i]. t2 == t1 && lrpocache [i]. t1 == t2 )
      {
         cmp = lrpocache [i]. cmp;
         if( cmp == COMP_LESS )
            return COMP_MORE;
         if( cmp == COMP_MORE )
            return COMP_LESS;
         return cmp;
      }
   }

   f1 = termspace [ t1 ];
   f2 = termspace [ t2 ];

   /* Handle the variable cases:           */

   if( f1 < 0 )
   {
      if( f2 < 0 )
      {
         if( f1 == f2 ) 
            cmp = COMP_EQUIV;
         else
            cmp = COMP_UNCOMPARABLE;
         goto returnvalue; 
      }
      else
      {
         if( symboloccurs( f1, t2 ))
            cmp = COMP_LESS;
         else
            cmp = COMP_UNCOMPARABLE;
         goto returnvalue;
      }
   }
   else
   {
      if( f2 < 0 )
      {
         if( symboloccurs( f2, t1 ))
            cmp = COMP_MORE;
         else
            cmp = COMP_UNCOMPARABLE;
         goto returnvalue;
      }
   }

   /* Both terms are non-variable.              */

#if 0
   printstring( &outputfile, "LRPO-nonvar( " );
   printterm( &outputfile, t1 );
   printstring( &outputfile, ", " );
   printterm( &outputfile, t2 );
   printstring( &outputfile, " )\n" );
#endif

   dir1 = operatorassoc [ f1 ];
   f1 = operatororder [ f1 ];
   dir2 = operatorassoc [ f2 ];
   f2 = operatororder [ f2 ];

   if( f1 < f2 )
      goto subtermsforless;
   if( f1 > f2 )
      goto subtermsformore;

   if( dir1 != dir2 )
   {
      cmp = COMP_UNCOMPARABLE;
      goto returnvalue;
   }

#if 0
   printstring( &outputfile, "operatorsequivalent: " );
   printterm( &outputfile, t1 );
   printstring( &outputfile, " == " );
   printterm( &outputfile, t2 );
   printstring( &outputfile, "\n" );
#endif

   if( dir1 < 0 )
      cmp = lr_lexext( t1, t2, lrpo ); 
   if( dir1 == 0 )
      cmp = multisetext( t1, t2, lrpo );
   if( dir1 > 0 ) 
      cmp = rl_lexext( t1, t2, lrpo );

   if( dir1 != 0 )
   {
      if( cmp == COMP_LESS )
         goto subtermsforless;
      if( cmp == COMP_MORE )
         goto subtermsformore;
      if( cmp == COMP_EQUIV )
         goto returnvalue;
      if( oneargumentgeq( t1, t2, lrpo ))
      {
         cmp = COMP_MORE;
         goto returnvalue;
      }

      if( oneargumentgeq( t2, t1, lrpo ))
      {
         cmp = COMP_LESS;
         goto returnvalue;
      }
      cmp = COMP_UNCOMPARABLE;
   }
   goto returnvalue;

subtermsforless:
   if( allargumentsless( t1, t2, lrpo ))
      cmp = COMP_LESS;
   else
   {
      if( oneargumentgeq( t1, t2, lrpo ))
         cmp = COMP_MORE;
      else
         cmp = COMP_UNCOMPARABLE;
   }
   goto returnvalue;

subtermsformore:
   if( allargumentsless( t2, t1, lrpo ))
      cmp = COMP_MORE;
   else
   {
      if( oneargumentgeq( t2, t1, lrpo ))
         cmp = COMP_LESS;
      else
         cmp = COMP_UNCOMPARABLE;
   }
   goto returnvalue;

returnvalue:
   if( freelrpocache >= MAXLRPOCACHE - 3 )
      errorwithoutline( "lrpo-cache is full" );

   lrpocache [ freelrpocache ]. t1 = t1;
   lrpocache [ freelrpocache ]. t2 = t2;
   lrpocache [ freelrpocache ]. cmp = cmp;
   freelrpocache ++ ;

   return cmp;
}



/**********************************************************************/
/*         Compare two terms under the termordering.                  */
/**********************************************************************/


int termcompare( term1, term2 )
int term1;
int term2;
{
   int d1, d2;
   int e;
   int te;

#if 0
   printstring( &outputfile, "termcompare " );
   printterm( &outputfile, term1 );
   printstring( &outputfile, " to  " );
   printterm( &outputfile, term2 );
   printstring( &outputfile, "\n\n" );
#endif

   if( termordering < 0 )
   {
      e = term1;
      term1 = term2;
      term2 = e;

      te = -termordering;
   }
   else
      te = termordering;

   switch( te )
   {
   case ORD_NONE: 
      return COMP_UNCOMPARABLE;
   case ORD_DEPTH:
      d1 = termdepth( term1 );
      d2 = termdepth( term2 );
      goto d1d2;
   case ORD_VARDEPTH:
      d1 = termvardepth( term1 );
      d2 = termvardepth( term2 );
      goto d1d2;
   case ORD_COMPL:
      d1 = termcompl( term1 );
      d2 = termcompl( term2 );
      goto d1d2;
   case ORD_NRVARS:
      d1 = termnrvars( term1 );
      d2 = termnrvars( term2 );
      goto d1d2; 
   case ORD_VD:
      return ord_d( term1, term2 );
   case ORD_KBO:
      return kbo( term1, term2 );
   case ORD_LRPO:
      freelrpocache = 0;
      return lrpo( term1, term2 );
   default:
      printf( " => %d ", te );
      errorwithoutline( "unknown order type for terms");
   }

d1d2:
   if( d1 < d2 ) return COMP_LESS;
   if( d1 > d2 ) return COMP_MORE;
   return COMP_EQUIV;
}
   

/* Derive a literal order from the term order, first sorting on the    */
/* predicate symbols, then on the arguments.                           */

static int exttermorderpa( lit1, lit2 )
int lit1;
int lit2;
{
   int op1, op2;
   int ord1, ord2;
   int dir;

#if 0 
   printstring( &outputfile, "exttermorderpa: ");
   printterm( &outputfile, lit1 );
   printstring( &outputfile, " == " );
   printterm( &outputfile, lit2 );
   printstring( &outputfile, " ?\n" );
#endif

   op1 = termspace [ lit1 ];
   op2 = termspace [ lit2 ];         

   ord1 = operatororder [ op1 ];
   ord2 = operatororder [ op2 ];
        
   if( ord1 == ord2 )
   {
     
      dir = operatorassoc [ op1 ];
      if( dir != operatorassoc [ op2 ] )
         return COMP_EQUIV;

      if( dir < 0 )
         return lr_lexext( lit1, lit2, termcompare );
      if( dir == 0 )
         return multisetext( lit1, lit2, termcompare );
      if( dir > 0 )
         return rl_lexext( lit1, lit2, termcompare );
   }

   if( ord1 < ord2 )
      return COMP_LESS;
   else
      return COMP_MORE;
}


/* Derive a literal order from the term order, first sorting on   */
/* the arguments, then on the predicate symbol.                   */
/* To my surprise sorting from left to right is useful, though    */
/* it does not sort orders.                                       */


static int exttermorderap( lit1, lit2 )
int lit1;
int lit2;
{
   int op1, op2;
   int ord1, ord2;
   int dir;
   int cmp;

#if 0 
   printstring( &outputfile, "exttermorderap: ");
   printterm( &outputfile, lit1 );
   printstring( &outputfile, " == " );
   printterm( &outputfile, lit2 );
   printstring( &outputfile, " ?\n" );
#endif

   op1 = termspace [ lit1 ];
   op2 = termspace [ lit2 ];         

   ord1 = operatororder [ op1 ];
   ord2 = operatororder [ op2 ];
       
   dir = operatorassoc [ op1 ];
   if( dir != operatorassoc [ op2 ] )
      return COMP_EQUIV;

#if 0   
      if( operatorassoc [ op1 ] == 0 && operatorassoc [ op2 ] != 0 ) 
         return COMP_LESS;
      if( operatorassoc [ op1 ] != 0 && operatorassoc [ op2 ] == 0 )
         return COMP_MORE;
      if( operatorassoc [ op1 ] < operatorassoc [ op2 ] )
         return COMP_LESS;
      if( operatorassoc [ op1 ] > operatorassoc [ op2 ] )
         return COMP_MORE;
      internalproblem( "you should not see this message" );
#endif

   if( dir < 0 )
      cmp = lr_lexext( lit1, lit2, termcompare );
   if( dir == 0 )
      cmp = multisetext( lit1, lit2, termcompare );
   if( dir > 0 )
      cmp = rl_lexext( lit1, lit2, termcompare );

   if( cmp != COMP_EQUIV ) 
      return cmp;

   if( ord1 < ord2 )
      return COMP_LESS;
   if( ord1 > ord2 ) 
      return COMP_MORE;
   return COMP_EQUIV;
}


 

/* compare t1 and t2 in a liftable order. t1 and t2 should be atoms. */ 

static int liftord( t1, t2)
int t1, t2;
{
   int f1, f2;
   int ord1, ord2;
   int dir;

#if 0
   printstring( &outputfile, "liftord: ");
   printterm( &outputfile, t1 );   
   printstring( &outputfile, " ; " );
   printterm( &outputfile, t2 );
   printstring( &outputfile, "\n" );
#endif

   f1 = termspace [ t1 ];
   f2 = termspace [ t2 ];

   if( f1 < 0 )
   {
      if( f2 < 0 )
      {
         if( f1 == f2 )
            return COMP_EQUIV;
         else
            return COMP_UNCOMPARABLE;
      }
      else
         return COMP_UNCOMPARABLE;
   }
   else
   {
      if( f2 < 0 )
         return COMP_UNCOMPARABLE;
   }

   ord1 = operatororder [ f1 ];
   ord2 = operatororder [ f2 ]; 

   if( ord1 == ord2 )
   {
      dir = operatorassoc [ f1 ];
      if( dir != operatorassoc [ f2 ] )
         return COMP_EQUIV;

      if( dir < 0 )
         return lr_lexext( t1, t2, liftord );
      if( dir == 0 )
         return multisetext( t1, t2, liftord );
      if( dir > 0 )
         return rl_lexext( t1, t2, liftord );
   }
   if( ord1 < ord2 )
      return COMP_LESS;
   else
      return COMP_MORE;
}


/* Compare t1 and t2 in an non-liftable order.                */

static int nonliftord( t1, t2 )
int t1, t2;
{
   int t1end; 
   int f1, f2;
   int i;

   t1end = termends [ t1 ];
   freevarprofile = 0;

   while( t1 < t1end )
   {

      f1 = termspace [ t1 ++ ] ;
      f2 = termspace [ t2 ++ ] ;
      
      if( f1 < 0 )
      {
         if( f2 < 0 ) 
         {
            for( i = 0; i < freevarprofile; i ++ ) 
            {
               if( varprofile [i]. var1 == f1 )
               {
                  if( varprofile [i]. var2 == f2 ) 
                     goto stillequal;
                  else
                     return COMP_LESS;
               }
               else
               {
                  if( varprofile [i]. var2 == f2 ) 
                     return COMP_MORE;
               }
            }
            varprofile [ freevarprofile ]. var1 = f1;
            varprofile [ freevarprofile ]. var2 = f2;
            freevarprofile ++ ; 
            goto stillequal;
         }
         else
         {
            return COMP_MORE;
         }
      }
      else
      {
         if( f2 < 0 ) 
         {
            return COMP_LESS;
         }
         else
         {
            if( f1 == f2 ) goto stillequal;
            f1 = operatororder [ f1 ];
            f2 = operatororder [ f2 ];
            if( f1 < f2 ) 
               return COMP_LESS;
            if( f1 > f2 ) 
               return COMP_MORE;
            return COMP_EQUIV;
         }
      }
   stillequal:
      ; 
   }
   return COMP_EQUIV;
}


/* Order which should be good with the splitting rule. */
/* It compares only literals of the form a, !b.        */

static int splitord( term1, term2 )
int term1;
int term2;
{
   int t1, t2;
   int pol1, pol2;
   int op1, op2;

   pol1 = termspace [ term1 ];
   pol2 = termspace [ term2 ];
   if( pol1 == OP_NOT )
      t1 = term1 + 1;
   else
      t1 = term1;

   if( pol2 == OP_NOT ) 
      t2 = term2 + 1;
   else
      t2 = term2;

   op1 = termspace [ t1 ];
   op2 = termspace [ t2 ];

   if( arities [ op1 ] != 0 && arities [ op2 ] != 0 )
      return COMP_EQUIV; 
   if( arities [ op1 ] != 0 )
   {
      if( pol2 == OP_NOT ) 
         return COMP_LESS;
      else
         return COMP_MORE;    
   }
   if( arities [ op2 ] != 0 ) 
   {
      if( pol1 == OP_NOT ) 
         return COMP_MORE;
      else
         return COMP_LESS;
   }   

   if( pol1 == OP_NOT && pol2 != OP_NOT ) return COMP_MORE;
   if( pol1 != OP_NOT && pol2 == OP_NOT ) return COMP_LESS;

   if( operatororder [ op1 ] < operatororder [ op2 ] )
      return COMP_LESS;
   if( operatororder [ op1 ] > operatororder [ op2 ] ) 
      return COMP_MORE;
   return COMP_EQUIV;
}

 
/*****************************************************************/   
/*            compare t1 with t2 under ordtype.                  */
/*****************************************************************/


static int compareord( t1, t2, ordtype )
int t1, t2;
int ordtype;
{
   int d1, d2;
  
#if 0
   printstring( &outputfile, "ord: ");
   printterm( &outputfile, t1 );
   printstring( &outputfile, ", ");
   printterm( &outputfile, t2 );
   printstring( &outputfile, " ?\n" );
#endif
 
   switch( ordtype )
   {
   case ORD_NONE:
      return COMP_UNCOMPARABLE;

   case ORD_DECISION:
      if( termspace [ t1 ] == OP_NOT ) 
         t1 ++ ;
      if( termspace [ t2 ] == OP_NOT ) 
         t2 ++ ; 
      d1 = termvardepth( t1 );
      d2 = termvardepth( t2 );
      if( d1 < d2 ) return COMP_LESS;
      if( d1 > d2 ) return COMP_MORE;
      d1 = termnrvars( t1 );
      d2 = termnrvars( t2 );
      if( d1 < d2 ) return COMP_LESS;
      if( d1 > d2 ) return COMP_MORE;
      return COMP_EQUIV;

   case ORD_EXTTERMORDPA:
      /* Extend the term order into an A-order,       */
      /* first sorting on predicates.                 */

      if( termspace [ t1 ] == OP_NOT )
         t1 ++ ;
      if( termspace [ t2 ] == OP_NOT ) 
         t2 ++ ;
      return exttermorderpa( t1, t2 );

   case ORD_EXTTERMORDAP:
      /* Extend the term order into an A-order,       */
      /* first sorting on arguments.                  */

      if( termspace [ t1 ] == OP_NOT ) 
         t1 ++ ;
      if( termspace [ t2 ] == OP_NOT ) 
         t2 ++ ;
      return exttermorderap( t1, t2 );

   case ORD_TERMORDERING:
      /* Use the term order as A-order.               */

      if( termspace [ t1 ] == OP_NOT ) 
         t1 ++ ;
      if( termspace [ t2 ] == OP_NOT ) 
         t2 ++ ;
      return termcompare( t1, t2 );
 
   case ORD_LIFTORD:
      /* A recursively computed, liftable A-order.    */

      if( termspace [ t1 ] == OP_NOT )
         t1 ++ ;
      if( termspace [ t2 ] == OP_NOT ) 
         t2 ++ ; 
      return liftord( t1, t2 );

   case ORD_NLO:
      /* A recursively computed, non-liftable A-order. */

      if( termspace [ t1 ] == OP_NOT ) 
         t1 ++ ; 
      if( termspace [ t2 ] == OP_NOT ) 
         t2 ++ ; 
      return nonliftord( t1, t2 ); 

   case ORD_SPLIT:
      return splitord( t1, t2 );

   default:
      printf("=> %d \n", ordtype );
      errorwithoutline( "unknown order type for atoms" ); 
   }
}


/************************************************************/
/* Compare lit1 with lit2 under the composition of the      */
/* orders. The order on terms, which is used for directing  */
/* equalities depends on the value of directeqs.            */
/************************************************************/

int literalcompare( lit1, lit2 )
int lit1;
int lit2;
{
   int i;
   int res;

#if 0
   printstring( &outputfile, "literalcompare " );
   printterm( &outputfile, lit1 );
   printstring( &outputfile, " to " );
   printterm( &outputfile, lit2 );
   printstring( &outputfile, "\n" );
#endif

   for( i = 0; i < nrlitorderings; i ++ )
   {
      res = compareord( lit1, lit2, litorderings [i] );
      if( res != COMP_EQUIV ) 
         return res;
   }
   return COMP_EQUIV;
}
  


/******************************************************************/
/*       Selection functions.                                     */
/******************************************************************/


/* negselect selects a literal for hyperresolution.       */
/* We don't know what is the best strategy and we select  */
/* without IQ.                                            */

int negselect( lit1, lit2 )
int lit1;
int lit2;
{
   int lit1end;
   int f1, f2;

#if 0
   printstring( &outputfile, "negselect: ");
   printterm( &outputfile, lit1 );
   printstring( &outputfile, ", " );
   printterm( &outputfile, lit2 );
   printstring( &outputfile, "\n" );
#endif

   f1 = termspace [ lit1 ];
   f2 = termspace [ lit2 ];

   if( f1 == OP_NOT && f2 != OP_NOT ) return COMP_MORE;
   if( f1 != OP_NOT && f2 == OP_NOT ) return COMP_LESS;
   if( f1 != OP_NOT && f2 != OP_NOT ) return COMP_EQUIV;

   lit1 ++ ;
   lit2 ++ ;
   f1 = termspace [ lit1 ];
   f2 = termspace [ lit2 ];
 
   lit1end = termends [ lit1 ];
   while( lit1 < lit1end )
   {
      f1 = termspace [ lit1 ++ ];
      f2 = termspace [ lit2 ++ ];
      
      if( f1 < f2 )                /* Reversing this makes bliksem */
         return COMP_MORE;         /* much worse. Why?             */
      if( f1 > f2 )
         return COMP_LESS;
   }
   return COMP_EQUIV;
}


/* posselect selects a literal for hyperresolution.       */
/* We don't know what is the best strategy and we select  */
/* without IQ.                                            */

int posselect( lit1, lit2 )
int lit1;
int lit2;
{
   int lit1end;
   int f1, f2;

#if 0
   printstring( &outputfile, "posselect: ");
   printterm( &outputfile, lit1 );
   printstring( &outputfile, ", ");
   printterm( &outputfile, lit2 );
   printstring( &outputfile, "\n");
#endif

   if( f1 == OP_NOT && f2 != OP_NOT ) return COMP_LESS;
   if( f1 != OP_NOT && f2 == OP_NOT ) return COMP_MORE;
   if( f1 == OP_NOT && f2 == OP_NOT ) return COMP_EQUIV;

   f1 = termspace [ lit1 ];
   f2 = termspace [ lit2 ];

   lit1end = termends [ lit1 ];
   while( lit1 < lit1end )
   {
      f1 = termspace [ lit1 ++ ];
      f2 = termspace [ lit2 ++ ];
      
      if( f1 < f2 )                /* Reversing this makes bliksem */
         return COMP_MORE;         /* much worse. Why?             */
      if( f1 > f2 )
         return COMP_LESS;
   }
   return COMP_EQUIV;
}


