

/**************************************************************/
/* resolution.h:  Datastructures related to resolution.       */
/* Bliksem 1.10                                               */
/**************************************************************/


/**********************************************************/
/* clauses contains some other things, but it is          */
/* called after its main use:                             */
/**********************************************************/


#define MAXCLAUSES   10000000
CLASS int clauses [ MAXCLAUSES ];
CLASS int freeclauses;


/*******************************************/
/*      Definition of a clause:            */
/*******************************************/


#define CLAUSE_NUMBER          0
   /* The first kept clause has number 0.                    */
#define CLAUSE_GENERATION      1
   /* The generation level of this clause. Initial clauses   */
   /* have generation 0.                                     */
#define CLAUSE_FILT1           2
#define CLAUSE_FILT2           3
#define CLAUSE_FILT3           4
#define CLAUSE_FILT4           5
#define CLAUSE_FILT5           6
#define CLAUSE_FILT6           7
#define CLAUSE_FILT7           8
#define CLAUSE_FILT8           9
   /* Subsumption bit filter.                               */
   /* 0 is used for an fpa that is there. 1 is used for     */
   /* an fpa that is not there. This is done for reasons    */
   /* of Hacking.                                           */
#define CLAUSE_NEXTINLIST     10
   /* The next clause in the clause list = listinuse/           */
   /* listpassive/listclauses/listunfactored/listproof.         */
#define CLAUSE_DERIVATION     11
   /* Reference to the derivation of the clause, see below for  */
   /* the structure.                                            */
#define CLAUSE_WEIGHT         12
   /* Total weight of the clause. Used for selection. */ 
#define CLAUSE_DEPTH          13 
   /* The depth of the clause.                        */ 
#define CLAUSE_LENGTH         14
   /* The length of the clause.                       */
#define CLAUSE_NRVARS         15
   /* Number of distinct variables in the clause.     */
#define CLAUSE_LITERALS       16
   /* Reference to the first literal.                 */ 
#define CLAUSE_MAXLITS        17
   /* Reference to the first maximal literal.         */ 
#define CLAUSE_END            18
   /* Reference to the end of the clause.             */

#define CLAUSE_REST           19
   /* Start of the rest of the clause.                */


/* So a clause has the following structure:                     */
/* 1. The header, on clauses [ clause ] ...                     */
/*                   clauses [ clause + CLAUSE_REST ].          */
/* 2. The derivation, on clauses [ clause + CLAUSE_DERIVATION ] */
/*               ...  clauses [ clause + CLAUSE_LITERALS ].     */
/* 3. The non-maximal literals, on                              */
/*             clauses [ clause + CLAUSE_LITERALS ] ...         */
/*             clauses [ clause + CLAUSE_MAXLITS ].             */
/* 4. The maximal literals, on                                  */
/*             clauses [ clause + CLAUSE_MAXLITS ] ...          */
/*             clauses [ clause + CLAUSE_END ].                 */


/* The derivation has the following form:                        */
/* RULE + Parents + ( PROC + parents )^* .                       */
/* So one rule + its parents, followed by 0 or                   */
/* postprocessings.                                              */


/* Binary Derivation Rules:   */

#define RULE_RES            (-100) 
   /* Ordered binary resolution.     (R)       */
#define RULE_PARAMOD        (-101) 
   /* Ordered binary paramodulation. (P)       */

/* Unary Derivation Rules:     */

#define RULE_EQREFL         (-200) 
   /* Equality reflexivity.          (Q)       */
#define RULE_EQFACT         (-201)
   /* Equality factoring.            (E)       */
#define RULE_FACTOR         (-202) 
   /* Binary factoring.              (F)       */
#define RULE_RESIMP         (-203)
   /* Resimplification.              (S)       */

/* Nullary Derivation Rules:   */

#define RULE_INITIAL        (-300)  
   /* Initial clauses.               (I)       */


#define SIMP_SPLITPOS       (-400)
   /* (p)                                      */
#define SIMP_SPLITNEG       (-401) 
   /* (n)                                      */
#define SIMP_DEMOD          (-402) 
   /* (d)                                      */
#define SIMP_RES            (-403) 
   /* (r)                                      */
#define SIMP_FACTOR         (-404)
   /* (f)                                      */
#define SIMP_EQREFL         (-405)   
   /* (q)                                      */
#define SIMP_SELF           (-406)
   /* (s)                                      */

 
/* Will keep the proof states in later versions. */

#define MAXPROOFSTATE 20
CLASS struct
{
   int usebinary;
   int translation;   /* The translation that was used. */
}
  proofstates [ MAXPROOFSTATE ];
CLASS int nrproofstates;
CLASS int currentproofstate;
 

/*******************************************************/
/*   Variables that control the deduction process:     */
/*******************************************************/

/* The binary derivation rules. They are applied when a clause is   */
/* moved from clauses to inuse. If 1, the rule is used.             */

CLASS int useres;
CLASS int useparamod;

/* The unary derivation rules. They are applied when a clause is moved  */
/* from unfactored to clauses. If 1, the rule is used.                  */

CLASS int useeqrefl;
CLASS int useeqfact;
CLASS int usefactor;


/* The simplification rules. They are applied when a clause is generated,  */
/* before it enters unfactored. The numbers specify the maximal            */
/* length that a simplifier can have. Setting a length < 0 means that the  */
/* simplification rule is not used.                                        */
/* Simplification rules are also applied on the given clause at the        */
/* moment that it is selected.                                             */
 
CLASS int usesimpsplitting;
CLASS int usesimpdemod;
CLASS int usesimpres;

/* It is possible to redo the simplification rules. This enables a         */
/* limited from of backward simplification.                                */

CLASS int resimpinuse, resimpinusecnt;
   /* Specifies how often clauses in inuse are resimplified.               */
CLASS int resimpclauses, resimpclausescnt;
   /* Specifies how often clauses in clauses are resimplified.             */
CLASS int showstatuscnt;
   /* Counter for showing intermediate status.                             */


/* Subsumption and tautology elimination cannot be switched off, at least  */
/* not here. They are also simplification rules. Tautology elimination is  */
/* not used as a resimplification rule, for obvious reasons.               */

#define SUBS_STANDARD    0
#define SUBS_SNL1        1
#define SUBS_EQREWR      2

CLASS int substype;
   /* If SUBS_STANDARD, then the standard type of subsumption.         */
   /* If SUBS_SNL1, then literals must be renamed by the substitution. */
   /* If SUBS_EQREWR then t1 = t2, t1 ==> t2 subsume                   */
   /* t1 = t2, t2 = t1, t1 ==> t2, t2 ==> t1.                          */

CLASS int backwardsubs; 
   /* This implements simple clausewise backward subsumption.          */
   /* If 0, then there is no backwardsubsumption, except through       */
   /* reprocessing of clauses.                                         */
   /* If 1, then a clause is allowed to backwardsubsume the other      */
   /* clauses inuse at selection time.                                 */
   /* If 2, a clause is allowed to backwardsubsume all other clauses,  */
   /* when it is kept. This option cannot be set by the user. Sorry!   */

CLASS int selectoldest, selectoldestcnt;
   /* Specifies how often the oldest clause is selected. */
   /* If selectoldest = 10, bliksem will select 9 times  */
   /* the shortest, and then 1 time the oldest.          */

CLASS int operatorweights [ MAXOPERATORS ];
   /* Weights are used only for selection. Not for sorting.     */


#define MAXLITORDERINGS     10
CLASS int litorderings [ MAXLITORDERINGS ];
CLASS int nrlitorderings;
   /* The complete literal ordering is the composition of these orderings. */
   /* Some literal orderings depend on a term ordering, in that case       */
   /* termordering must be defined.                                        */
CLASS int termordering;
   /* The ordering used for directing equalities. Some literal orderings   */
   /* are defined as extensions of a termordering.                         */

CLASS int litapriori;
   /* If 1, then clauses are sorted when they are generated, using       */
   /* litorderings.                                                      */
CLASS int termapriori;
   /* If 1, then equalities are directed and undirected when they are    */
   /* generated, using termordering.                                     */
CLASS int litaposteriori;
   /* If 1, then litorderings will be applied a posteriori.              */
CLASS int termaposteriori;
   /* If 1, then termordering will be applied a posteriori               */
CLASS int demodaposteriori;
   /* If 1, then demodulation will be attempted with a posteriori        */
   /* direction of the equalities.                                       */
CLASS int ordereqreflfact;
   /* If 1, then equality reflexivity and factoring are ordererd.        */
   /* How they are ordered depends on litapriori, litaposteriori.        */

CLASS int operatororder [ MAXOPERATORS ]; 
   /* In general higher numbers are more preferred.              */
CLASS int operatorassoc [ MAXOPERATORS ];
   /* Used by some orders. This makes it possible to direct      */
   /* associativity axioms.                                      */
   /* A negative number means preference for rightassociativity. */
   /* Zero means that there is no associativity preference.      */
   /* A positive number means preference for leftassociativity.  */

CLASS int litselect;
   /* Sets a selection function. Selection is done after         */
   /* in order to allow the selection function to overrule the   */
   /* ordering.                                                  */

CLASS int operatorselect [ MAXOPERATORS ];
   /* Whether or not the operator is a candidate for negative    */
   /* selection by ORD_NEGSELECT. Type predicates should not     */
   /* be selected, because this leads to enumeration of the      */
   /* Herbrand base.                                             */

CLASS int operatornoparamod [ MAXOPERATORS ];
   /* If operatornoparamod [ op ] == 1, then there will be no    */
   /* paramodulation into the arguments of a functional term     */
   /* built with op. This should be set for Skolem functions,    */
   /* symbols resulting from the structural transformation, and  */
   /* type predicates.                                           */


CLASS int maxweight;
CLASS int maxdepth;
CLASS int maxlength;
CLASS int maxnrvars;
   /* If any of these value exceeds its limit, the clause is not kept,  */
   /* unless when the generation level is less than or equal to         */
   /* excuselevel. The generation level of initial clauses is 0.        */

CLASS int excuselevel;
   /* The complexity restrictions do not apply on clauses up to the     */
   /* generation level specified by excuselevel.                        */
   /* Initial clauses have generation 0.                                */
CLASS int increasemaxweight; 
   /* If 1, then maxweight will be increased by 1 after a failed proof  */
   /* attempt.                                                          */

CLASS int maxselected;
   /* Bliksem stops when maxselected reaches its limit.                 */
CLASS int maxclauses;
   /* If maxclauses reaches it limit, then Bliksem starts deleting      */
   /* heavy clauses.                                                    */

CLASS int showgenerated;
CLASS int showkept;
CLASS int showselected;
CLASS int showdeleted;    /* by backward subsumption, or replacement. */ 
CLASS int showresimp;
   /* Determine which clauses are printed.                            */
CLASS int showstatus;
   /* This determines how often a status report is printed.           */

/*********************************************************/
/*  Variables used during the deduction process:         */
/*********************************************************/


#define MAXNRVARS        100 
   /* The maximal number of variables allowed in one clause. */ 
#define MAXCLAUSELEN      50 
   /* Maximal length of a clause.                            */



CLASS int nrgenerated;
   /* Total number of generated clauses.          */
   /* Includes the clauses that are not kept.     */ 
CLASS int nrkept;
   /* The number to be given to the next clause.  */
CLASS int nrselected;
   /* How many clauses have been selected.        */
CLASS int nrdeleted;
   /* How many clauses were deleted, after they   */
   /* were kept.                                  */
CLASS int nrdeletedinuse;
   /* How many clauses were deleted from inuse.   */

/* Is this still useful?                          */
CLASS int subsentry;
   /* How often the subsumption test was entered. */
CLASS int subspropmatch;
   /* How often the literals could be matched     */
   /* individually, but without checking          */
   /* variable consistency.                       */
CLASS int subslitmatch;
   /* How often the literals individually could   */
   /* be matched.                                 */
CLASS int subsfullsubs;
   /* How often the full subsumption test was     */
   /* entered.                                    */


/* clause lists built by NEXTINLIST:                             */

#define EMPTYCLAUSELIST     (-1) 

CLASS int listinuse;
CLASS int listpassive;
CLASS int listclauses;
CLASS int listunfactored; 
CLASS int listproof;
CLASS int listdeleted;
CLASS int listdeletedinuse; 
   /* We distinghuish deleted clauses that have been used from     */
   /* deleted clauses that have not been used. The latter can be   */
   /* physically deleted, when memory is getting full.             */



#define MAXLISTSCLAUSES       40 
CLASS struct
{   
   int *lightest;
   int *oldest;     
   int clauselist;  /* A list of clauses, connected by            */
                    /* NEXTINLIST.                                */

}
   lllistclauses [ MAXLISTSCLAUSES ];

CLASS int listclausescnt;
   /* Determines into which of listclauses [ ], the next insertion */
   /* will take place.                                             */

   /* To be used in the future, to reduce the time of selecting a  */
   /* a clause.                                                    */

CLASS int treeforwdemod; 
   /* Contains equalities that can be used as demodulator.         */
CLASS int treeforwrespos;
   /* Contains positive literals for resolution simplification.    */
CLASS int treeforwresneg; 
   /* Contains negative literals for resolution simplification.    */
   /* Negation signs are not stored.                               */
CLASS int treeforwsubs;
   /* Contains most unlikely literal of a clause for subsumption   */
   /* test.                                                        */

CLASS int incomplete;
   /* Bliksem sets incomplete = 1, when it is certain that the     */
   /* strategy that it uses is incomplete. This happens for        */
   /* example, when a clause is discarded by its weight, or when   */
   /* an upperbound on kept/selected clauses is reached.           */
   /* In the case that incomplete = 0, this means nothing.         */



/********************************************************************/
/*               Orders:                                            */
/********************************************************************/


/* The types of orders:                                   */

#define ORD_NONE                         0
   /* Everything is uncomparable to everything.           */

#define ORD_DEPTH                        1
   /* Prefer deeper literals. The negation sign does      */
   /* not contribute to the depth.                        */

#define ORD_VARDEPTH                     2
   /* Prefer literals with greater vardepth,              */
   /* The negation sign does not contribute to the        */
   /* vardepth.                                           */

#define ORD_COMPL                        3
   /* Prefer more complex literals, using                 */
   /* operatororder []. Variables have complexity 1.      */

#define ORD_NRVARS                       4
   /* Prefer literals with more different variables.      */

#define ORD_VD                           5
   /* A < B if every variable in A occurs strictly deeper */
   /* in B, and Depth(A) < Depth(B).                      */
   /* Originates from FLTZ93.                             */
   /* The negation symbol is ignored.                     */

#define ORD_DECISION                     6
   /* Order for decidable classes. A < B iff                */
   /* Varnr( A ) \subset Varnr( B ) or                      */
   /* Vardepth(A) < Vardepth( B ).                          */
#define ORD_KBO                          7
   /* Knuth-Bendix order.                                   */
   /* A < B iff A is strictly less complex than B, and      */
   /* every variable in A occurs at least as often in B.    */
   /* ORD_KBO uses a complexity notion based on             */
   /* operatororder, or gives equal complexity to the       */
   /* symbols. I currently don't know the best way.         */
   /* operatororder determines the complexity of the        */
   /* individual operators. Variables have complexity 1.    */
   /* operatorassoc determines the direction of comparison  */
   /* of subterms. Putting operatorassoc == 0 is not a      */
   /* good idea, < 0 means lr, > 0 means rl.                */

#define ORD_LRPO                         8
   /* Lexicographic path order, with recursive path order.  */
   /* It uses operatororder [] to sort the operatorsymbols. */
   /* operatorassoc [] is used to decide what to do with    */
   /* the arguments. < 0 means that arguments are compared  */
   /* left/right. > 0 means that arguments are compared     */
   /* right/left. == 0, means that arguments are compared   */
   /* a multiset fashion.                                   */
      
#define ORD_EXTTERMORDPA                 9
#define ORD_EXTTERMORDAP                10
   /* Extend the term order to a literal order, using       */
   /* operatorassoc [], and operatororder [].               */
   /* These can be combined with every term order.          */
   /* PA sorts first on the predicate symbol, then on the   */
   /* arguments. AP sorts first on the arguments, then on   */
   /* the predicate symbol. Sorting on the arguments is     */
   /* always done in a multiset fashion. Any alternative    */
   /* would make equality literals uncomparable to the      */
   /* the other literals. This is certainly not what one    */
   /* wants.                                                */

#define ORD_TERMORDERING                11
   /* Make an A-order treating the atoms as terms, and      */
   /* using the term order to sort them.                    */

#define ORD_LIFTORD                     12
   /* A recursive, liftable A-order, uses operatororder,    */
   /* and operatorassoc. Subterms are compared, dependent   */
   /* on operatorassoc.                                     */

#define ORD_NLO                         13
   /* A non-liftable A-order, uses operatororder.           */

#define ORD_SPLIT                       14
   /* Order to be used in combination with the              */
   /* splitting rule. It compares A with B if one of A, B   */
   /* is a propositional literal. It puts negative,         */
   /* propositional literals maximal. It puts positive      */
   /* propositional literals minimal.                       */
   /* The rest is equivalent.                               */


/**************************************************************/
/*     Literal Selection Functions.                           */
/**************************************************************/


#define LITSEL_NONE                        0
   /* Don't select literals.                                 */
#define LITSEL_POS                         1
   /* Always select a positive literal.                      */
#define LITSEL_NEG                         2
   /* Always select a negative literal.                      */
#define LITSEL_POSORD                      3
   /* Select a positive literal from the maximal literals.   */
#define LITSEL_NEGORD                      4
   /* Select a negative literal from the maximal literals.   */


/*************************************************/
/* Possible values of comparisons.               */
/* We see an order as a relation R satisfying    */
/* the following conditions:                     */
/* [x] R(x,x).                                   */
/* [x,y,z] R(x,y) & R(y,z) -> R(x,z).            */
/* So, in fact as a pre-order.                   */
/*************************************************/


#define COMP_LESS                (-1)
   /* R(x,y) & ! R(y,x).   */
#define COMP_EQUIV                0
   /* R(x,y) & R(y,x).     */
#define COMP_MORE                 1
   /* ! R(x,y) & R(y,x).   */ 
#define COMP_UNCOMPARABLE       100
   /* ! R(x,y) & ! R(y,x). */ 



/********************************************************************/
/* Variables controlling the input and the conversion.              */
/********************************************************************/

/* Used for storing the input:                                      */

CLASS int startofinput, endofinput;
   /* termspace [ startofinput .. endofinput ] contains the input   */
   /* formulae. If some input formulae are to be used as goals,     */
   /* then these should be last.                                    */

#define MAXTRANSLATIONS 10
CLASS struct
{
   int fullconv;                        /* FULLCONV used.                  */
   int from1, from2;                    /* termspace [ from1 .. from2 ]    */
                                        /* specifies from where the        */
                                        /* transformation is obtained.     */
   int startofresult, endofresult;      /* termspace [ sor .. eor ]        */
                                        /* contains the result.            */
   int startofjust, endofjust;          /* justifications [ soj .. eoj ]   */
                                        /* contains the specification of   */
                                        /* the transformation, in the case */
                                        /* that totalproof = 1.            */
                                        
}
   translations [ MAXTRANSLATIONS ];
CLASS int nrtranslations;

CLASS int nrgoals;        /* The last nrgoals clauses are treated        */
                          /* as goals.                                   */


/*************************************************************************/
/*    If totalproof = 1, then Bliksem outputs a full report on the       */
/*    conversion, and on the deduction process.                          */
/*************************************************************************/


CLASS int totalproof;


/****************************************************************/
/*   Conversion rules that can be used in transformation to     */
/*   clausal normal form.                                       */
/****************************************************************/


#define CONV_NNF                 1
   /* Produce the negation normal form.                   */

#define CONV_SKOLEMFREE          2
   /* Skolemize using only free variables.                */
#define CONV_SKOLEM              3
   /* Skolemize using all variables.                      */
#define CONV_SKOLEMNG            4
   /* Skolemization using Nonnengart's optimization.      */

#define CONV_FACTOR              5
   /* Factor into first order clauses.                    */

#define CONV_STRUCTNA1           6             
#define CONV_STRUCTNA2           7
   /* Structural removal of double arrows.                */
   /* The difference between 1 and 2 is that 1 makes      */
   /* a total removal, and that 2 has some tolerance.     */

#define CONV_STRUCTFACT1         8
#define CONV_STRUCTFACT2         9
   /* Double sided structural replacement of subterms     */
   /* that would cause large factorings of &'s over       */
   /* |'s.                                                */

#define CONV_STRUCTFACTOS1      10
#define CONV_STRUCTFACTOS2      11
   /* One sided structural replacement of subterms that   */
   /* would cause large factorings of &'s over |'s.       */

#define CONV_STRUCTGUARD        12 
   /* A special structural translation for the guarded    */
   /* fragment.                                           */
#define CONV_STRUCT2VAR         13
   /* A special structural translation for the 2-variable */
   /* fragment.                                           */

#define CONV_SIMPANTIPRENEX     14
   /* Full antiprenexing. It also permutes &'s.           */
#define CONV_SIMPQUANTFACT      15
   /* Try to factor quantifiers inward when they do not   */
   /* occur in one of the members of a binary operator.   */
   /* This (hopefully) produces smaller Skolem terms,     */
   /* may improve the effect of CONV_SKOLEMNG.            */


#define MAXEQUIVDIFF                (50.0) 
   /* Controls STRUCTNA2.                                 */
#define MAXANDINOR                  (50.0) 
   /* Controls STRUCTFACT2 and STRUCTFACTOS2.             */


/**********************************************************/
/*       Full conversions:                                */
/**********************************************************/

#define FULLCONV_NONE          0
#define FULLCONV_STANDARD      1
#define FULLCONV_GUARDED       2
#define FULLCONV_STRUCT        3
#define FULLCONV_STRUCTOS      4
#define FULLCONV_2VAR          5


/**********************************************************/
/*    Substitutions and matchings:                        */
/**********************************************************/


#define MAXSUBST 1000 
CLASS struct
{
   int var; 
   int varorig; 
   int term; 
   int termorig;
} 
   subst [ MAXSUBST ];
CLASS int freesubst;



CLASS struct
{
   int var;
   int varorig;
}
   renaming [ MAXNRVARS ];
CLASS int freerenaming;



#define MAXMATCHING    1000
CLASS struct  
{
   int var; 
   int term; 
}
   matching [ MAXMATCHING ];
CLASS int freematching;


/*************************************************************/
/* Stores the literals of a new clause, and its parents.     */
/*************************************************************/


/* Literals of a fresh clause.      */
/* References into termspace:       */

#define MAXLITERALS ( 2 * MAXCLAUSELEN )
CLASS int literals [ MAXLITERALS ];
CLASS int nrliterals;


/* The parents of a fresh clause:   */

#define MAXPARENTS  1000 
CLASS int parents [ MAXPARENTS ];
CLASS int freeparents;



/*************************************************************/ 
/* Used to build discrimination trees. Discrimination trees  */
/* are stored in the array clauses.                          */
/* THIS IS TO BE REPLACED.                                   */
/*************************************************************/


#define DT_NIL   (-1) 
   /* Used for end of list.                                   */
#define DT_REC   (-2) 
   /* A discrimination tree record.                           */
   /* 0: DT_REC itself.                                       */
   /* 1: An operator ( >= 0 ), or a variable ( < 0 ).         */
   /* 2: A reference for when the operator/variable succeeds. */
   /* 3: A reference for when the operator/variable fails.    */
#define DT_SUCC  (-3) 
   /* A success node.                                         */
   /* 0: DT_SUCC itself.                                      */
   /* 1: A reference to another solution.                     */
   /* 2: A renaming substution, enumerating the real values   */
   /*    for the variables -1, -2, -3, ...                    */
   /* Position i can be recognized as the first >= 0 integer  */
   /* after 2.                                                */
   /* i: The clause that contains the term, index in clauses. */
   /* i+1: Index in clauses of the literal that contains the  */
   /*      term.                                              */
   /* i+2: Index in terms of the term. It is possible that    */
   /*      the term is a subterm of a literal.                */

   /* It is essential that no subtree is empty.               */



/******************************************************/
/* Used for storing details of the proof. Used in     */
/* proofcheck.c and in conversion.c                   */
/******************************************************/


#define MAXJUSTIFICATIONS    5000000
CLASS int justifications [ MAXJUSTIFICATIONS ];
CLASS int freejustifications; 


/*********************************************************************/
/* Conversions are stored as follows:                                */
/* The conversions CONV_NOARROW, CONV_NNF, CONF_FACTOR               */
/* are stored as pairs type << 8 ) | rule, relative position.        */
/* The conversions CONV_SKOLEMFREE, CONV_SKOLEM, CONV_SKOLEMNG       */
/* CONV_STRUCT, CONV_STRUCTOS, CONV_STRUCTFACT are stored as         */
/* type << 8, replacement formula terminated by -1, relative         */
/* position.                                                         */
/*********************************************************************/


